from typing import overload
from enum import Enum
import abc
import datetime
import typing

import QuantConnect
import QuantConnect.Data
import QuantConnect.Data.Market
import QuantConnect.Interfaces
import QuantConnect.Orders
import QuantConnect.Securities
import QuantConnect.Statistics
import System
import System.Collections.Generic


class PortfolioStatistics(System.Object):
    """The PortfolioStatistics class represents a set of statistics calculated from equity and benchmark samples"""

    @property
    def average_win_rate(self) -> float:
        """The average rate of return for winning trades"""
        ...

    @property.setter
    def average_win_rate(self, value: float) -> None:
        ...

    @property
    def average_loss_rate(self) -> float:
        """The average rate of return for losing trades"""
        ...

    @property.setter
    def average_loss_rate(self, value: float) -> None:
        ...

    @property
    def profit_loss_ratio(self) -> float:
        """The ratio of the average win rate to the average loss rate"""
        ...

    @property.setter
    def profit_loss_ratio(self, value: float) -> None:
        ...

    @property
    def win_rate(self) -> float:
        """The ratio of the number of winning trades to the total number of trades"""
        ...

    @property.setter
    def win_rate(self, value: float) -> None:
        ...

    @property
    def loss_rate(self) -> float:
        """The ratio of the number of losing trades to the total number of trades"""
        ...

    @property.setter
    def loss_rate(self, value: float) -> None:
        ...

    @property
    def expectancy(self) -> float:
        """The expected value of the rate of return"""
        ...

    @property.setter
    def expectancy(self, value: float) -> None:
        ...

    @property
    def start_equity(self) -> float:
        """Initial Equity Total Value"""
        ...

    @property.setter
    def start_equity(self, value: float) -> None:
        ...

    @property
    def end_equity(self) -> float:
        """Final Equity Total Value"""
        ...

    @property.setter
    def end_equity(self, value: float) -> None:
        ...

    @property
    def compounding_annual_return(self) -> float:
        """Annual compounded returns statistic based on the final-starting capital and years."""
        ...

    @property.setter
    def compounding_annual_return(self, value: float) -> None:
        ...

    @property
    def drawdown(self) -> float:
        """Drawdown maximum percentage."""
        ...

    @property.setter
    def drawdown(self, value: float) -> None:
        ...

    @property
    def total_net_profit(self) -> float:
        """The total net profit percentage."""
        ...

    @property.setter
    def total_net_profit(self, value: float) -> None:
        ...

    @property
    def sharpe_ratio(self) -> float:
        """Sharpe ratio with respect to risk free rate: measures excess of return per unit of risk."""
        ...

    @property.setter
    def sharpe_ratio(self, value: float) -> None:
        ...

    @property
    def probabilistic_sharpe_ratio(self) -> float:
        """
        Probabilistic Sharpe Ratio is a probability measure associated with the Sharpe ratio.
        It informs us of the probability that the estimated Sharpe ratio is greater than a chosen benchmark
        """
        ...

    @property.setter
    def probabilistic_sharpe_ratio(self, value: float) -> None:
        ...

    @property
    def sortino_ratio(self) -> float:
        """Sortino ratio with respect to risk free rate: measures excess of return per unit of downside risk."""
        ...

    @property.setter
    def sortino_ratio(self, value: float) -> None:
        ...

    @property
    def alpha(self) -> float:
        """Algorithm "Alpha" statistic - abnormal returns over the risk free rate and the relationshio (beta) with the benchmark returns."""
        ...

    @property.setter
    def alpha(self, value: float) -> None:
        ...

    @property
    def beta(self) -> float:
        """Algorithm "beta" statistic - the covariance between the algorithm and benchmark performance, divided by benchmark's variance"""
        ...

    @property.setter
    def beta(self, value: float) -> None:
        ...

    @property
    def annual_standard_deviation(self) -> float:
        """Annualized standard deviation"""
        ...

    @property.setter
    def annual_standard_deviation(self, value: float) -> None:
        ...

    @property
    def annual_variance(self) -> float:
        """Annualized variance statistic calculation using the daily performance variance and trading days per year."""
        ...

    @property.setter
    def annual_variance(self, value: float) -> None:
        ...

    @property
    def information_ratio(self) -> float:
        """Information ratio - risk adjusted return"""
        ...

    @property.setter
    def information_ratio(self, value: float) -> None:
        ...

    @property
    def tracking_error(self) -> float:
        """Tracking error volatility (TEV) statistic - a measure of how closely a portfolio follows the index to which it is benchmarked"""
        ...

    @property.setter
    def tracking_error(self, value: float) -> None:
        ...

    @property
    def treynor_ratio(self) -> float:
        """Treynor ratio statistic is a measurement of the returns earned in excess of that which could have been earned on an investment that has no diversifiable risk"""
        ...

    @property.setter
    def treynor_ratio(self, value: float) -> None:
        ...

    @property
    def portfolio_turnover(self) -> float:
        """The average Portfolio Turnover"""
        ...

    @property.setter
    def portfolio_turnover(self, value: float) -> None:
        ...

    @property
    def value_at_risk_99(self) -> float:
        """
        The 1-day VaR for the portfolio, using the Variance-covariance approach.
        Assumes a 99% confidence level, 1 year lookback period, and that the returns are normally distributed.
        """
        ...

    @property.setter
    def value_at_risk_99(self, value: float) -> None:
        ...

    @property
    def value_at_risk_95(self) -> float:
        """
        The 1-day VaR for the portfolio, using the Variance-covariance approach.
        Assumes a 95% confidence level, 1 year lookback period, and that the returns are normally distributed.
        """
        ...

    @property.setter
    def value_at_risk_95(self, value: float) -> None:
        ...

    @overload
    def __init__(self, profitLoss: System.Collections.Generic.SortedDictionary[datetime.datetime, float], equity: System.Collections.Generic.SortedDictionary[datetime.datetime, float], portfolioTurnover: System.Collections.Generic.SortedDictionary[datetime.datetime, float], listPerformance: System.Collections.Generic.List[float], listBenchmark: System.Collections.Generic.List[float], startingCapital: float, riskFreeInterestRateModel: QuantConnect.Data.IRiskFreeInterestRateModel, tradingDaysPerYear: int, winCount: typing.Optional[int] = None, lossCount: typing.Optional[int] = None) -> None:
        """
        Initializes a new instance of the PortfolioStatistics class
        
        :param profitLoss: Trade record of profits and losses
        :param equity: The list of daily equity values
        :param portfolioTurnover: The algorithm portfolio turnover
        :param listPerformance: The list of algorithm performance values
        :param listBenchmark: The list of benchmark values
        :param startingCapital: The algorithm starting capital
        :param riskFreeInterestRateModel: The risk free interest rate model to use
        :param tradingDaysPerYear: The number of trading days per year
        :param winCount: The number of wins, including ITM options with profitLoss less than 0. If this and  are null, they will be calculated from
        :param lossCount: The number of losses
        """
        ...

    @overload
    def __init__(self) -> None:
        """Initializes a new instance of the PortfolioStatistics class"""
        ...


class TradeDirection(Enum):
    """Direction of a trade"""

    LONG = 0
    """Long direction (0)"""

    SHORT = 1
    """Short direction (1)"""


class Trade(System.Object):
    """Represents a closed trade"""

    @property
    def symbol(self) -> QuantConnect.Symbol:
        """The symbol of the traded instrument"""
        ...

    @property.setter
    def symbol(self, value: QuantConnect.Symbol) -> None:
        ...

    @property
    def entry_time(self) -> datetime.datetime:
        """The date and time the trade was opened"""
        ...

    @property.setter
    def entry_time(self, value: datetime.datetime) -> None:
        ...

    @property
    def entry_price(self) -> float:
        """The price at which the trade was opened (or the average price if multiple entries)"""
        ...

    @property.setter
    def entry_price(self, value: float) -> None:
        ...

    @property
    def direction(self) -> QuantConnect.Statistics.TradeDirection:
        """The direction of the trade (Long or Short)"""
        ...

    @property.setter
    def direction(self, value: QuantConnect.Statistics.TradeDirection) -> None:
        ...

    @property
    def quantity(self) -> float:
        """The total unsigned quantity of the trade"""
        ...

    @property.setter
    def quantity(self, value: float) -> None:
        ...

    @property
    def exit_time(self) -> datetime.datetime:
        """The date and time the trade was closed"""
        ...

    @property.setter
    def exit_time(self, value: datetime.datetime) -> None:
        ...

    @property
    def exit_price(self) -> float:
        """The price at which the trade was closed (or the average price if multiple exits)"""
        ...

    @property.setter
    def exit_price(self, value: float) -> None:
        ...

    @property
    def profit_loss(self) -> float:
        """The gross profit/loss of the trade (as account currency)"""
        ...

    @property.setter
    def profit_loss(self, value: float) -> None:
        ...

    @property
    def total_fees(self) -> float:
        """The total fees associated with the trade (always positive value) (as account currency)"""
        ...

    @property.setter
    def total_fees(self, value: float) -> None:
        ...

    @property
    def mae(self) -> float:
        """The Maximum Adverse Excursion (as account currency)"""
        ...

    @property.setter
    def mae(self, value: float) -> None:
        ...

    @property
    def mfe(self) -> float:
        """The Maximum Favorable Excursion (as account currency)"""
        ...

    @property.setter
    def mfe(self, value: float) -> None:
        ...

    @property
    def duration(self) -> datetime.timedelta:
        """Returns the duration of the trade"""
        ...

    @property
    def end_trade_drawdown(self) -> float:
        """Returns the amount of profit given back before the trade was closed"""
        ...

    @property
    def is_win(self) -> bool:
        """Returns whether the trade was profitable (is a win) or not (a loss)"""
        ...

    @property.setter
    def is_win(self, value: bool) -> None:
        ...


class TradeStatistics(System.Object):
    """The TradeStatistics class represents a set of statistics calculated from a list of closed trades"""

    @property
    def start_date_time(self) -> typing.Optional[datetime.datetime]:
        """The entry date/time of the first trade"""
        ...

    @property.setter
    def start_date_time(self, value: typing.Optional[datetime.datetime]) -> None:
        ...

    @property
    def end_date_time(self) -> typing.Optional[datetime.datetime]:
        """The exit date/time of the last trade"""
        ...

    @property.setter
    def end_date_time(self, value: typing.Optional[datetime.datetime]) -> None:
        ...

    @property
    def total_number_of_trades(self) -> int:
        """The total number of trades"""
        ...

    @property.setter
    def total_number_of_trades(self, value: int) -> None:
        ...

    @property
    def number_of_winning_trades(self) -> int:
        """The total number of winning trades"""
        ...

    @property.setter
    def number_of_winning_trades(self, value: int) -> None:
        ...

    @property
    def number_of_losing_trades(self) -> int:
        """The total number of losing trades"""
        ...

    @property.setter
    def number_of_losing_trades(self, value: int) -> None:
        ...

    @property
    def total_profit_loss(self) -> float:
        """The total profit/loss for all trades (as symbol currency)"""
        ...

    @property.setter
    def total_profit_loss(self, value: float) -> None:
        ...

    @property
    def total_profit(self) -> float:
        """The total profit for all winning trades (as symbol currency)"""
        ...

    @property.setter
    def total_profit(self, value: float) -> None:
        ...

    @property
    def total_loss(self) -> float:
        """The total loss for all losing trades (as symbol currency)"""
        ...

    @property.setter
    def total_loss(self, value: float) -> None:
        ...

    @property
    def largest_profit(self) -> float:
        """The largest profit in a single trade (as symbol currency)"""
        ...

    @property.setter
    def largest_profit(self, value: float) -> None:
        ...

    @property
    def largest_loss(self) -> float:
        """The largest loss in a single trade (as symbol currency)"""
        ...

    @property.setter
    def largest_loss(self, value: float) -> None:
        ...

    @property
    def average_profit_loss(self) -> float:
        """The average profit/loss (a.k.a. Expectancy or Average Trade) for all trades (as symbol currency)"""
        ...

    @property.setter
    def average_profit_loss(self, value: float) -> None:
        ...

    @property
    def average_profit(self) -> float:
        """The average profit for all winning trades (as symbol currency)"""
        ...

    @property.setter
    def average_profit(self, value: float) -> None:
        ...

    @property
    def average_loss(self) -> float:
        """The average loss for all winning trades (as symbol currency)"""
        ...

    @property.setter
    def average_loss(self, value: float) -> None:
        ...

    @property
    def average_trade_duration(self) -> datetime.timedelta:
        """The average duration for all trades"""
        ...

    @property.setter
    def average_trade_duration(self, value: datetime.timedelta) -> None:
        ...

    @property
    def average_winning_trade_duration(self) -> datetime.timedelta:
        """The average duration for all winning trades"""
        ...

    @property.setter
    def average_winning_trade_duration(self, value: datetime.timedelta) -> None:
        ...

    @property
    def average_losing_trade_duration(self) -> datetime.timedelta:
        """The average duration for all losing trades"""
        ...

    @property.setter
    def average_losing_trade_duration(self, value: datetime.timedelta) -> None:
        ...

    @property
    def median_trade_duration(self) -> datetime.timedelta:
        """The median duration for all trades"""
        ...

    @property.setter
    def median_trade_duration(self, value: datetime.timedelta) -> None:
        ...

    @property
    def median_winning_trade_duration(self) -> datetime.timedelta:
        """The median duration for all winning trades"""
        ...

    @property.setter
    def median_winning_trade_duration(self, value: datetime.timedelta) -> None:
        ...

    @property
    def median_losing_trade_duration(self) -> datetime.timedelta:
        """The median duration for all losing trades"""
        ...

    @property.setter
    def median_losing_trade_duration(self, value: datetime.timedelta) -> None:
        ...

    @property
    def max_consecutive_winning_trades(self) -> int:
        """The maximum number of consecutive winning trades"""
        ...

    @property.setter
    def max_consecutive_winning_trades(self, value: int) -> None:
        ...

    @property
    def max_consecutive_losing_trades(self) -> int:
        """The maximum number of consecutive losing trades"""
        ...

    @property.setter
    def max_consecutive_losing_trades(self, value: int) -> None:
        ...

    @property
    def profit_loss_ratio(self) -> float:
        """The ratio of the average profit per trade to the average loss per trade"""
        ...

    @property.setter
    def profit_loss_ratio(self, value: float) -> None:
        ...

    @property
    def win_loss_ratio(self) -> float:
        """The ratio of the number of winning trades to the number of losing trades"""
        ...

    @property.setter
    def win_loss_ratio(self, value: float) -> None:
        ...

    @property
    def win_rate(self) -> float:
        """The ratio of the number of winning trades to the total number of trades"""
        ...

    @property.setter
    def win_rate(self, value: float) -> None:
        ...

    @property
    def loss_rate(self) -> float:
        """The ratio of the number of losing trades to the total number of trades"""
        ...

    @property.setter
    def loss_rate(self, value: float) -> None:
        ...

    @property
    def average_mae(self) -> float:
        """The average Maximum Adverse Excursion for all trades"""
        ...

    @property.setter
    def average_mae(self, value: float) -> None:
        ...

    @property
    def average_mfe(self) -> float:
        """The average Maximum Favorable Excursion for all trades"""
        ...

    @property.setter
    def average_mfe(self, value: float) -> None:
        ...

    @property
    def largest_mae(self) -> float:
        """The largest Maximum Adverse Excursion in a single trade (as symbol currency)"""
        ...

    @property.setter
    def largest_mae(self, value: float) -> None:
        ...

    @property
    def largest_mfe(self) -> float:
        """The largest Maximum Favorable Excursion in a single trade (as symbol currency)"""
        ...

    @property.setter
    def largest_mfe(self, value: float) -> None:
        ...

    @property
    def maximum_closed_trade_drawdown(self) -> float:
        """The maximum closed-trade drawdown for all trades (as symbol currency)"""
        ...

    @property.setter
    def maximum_closed_trade_drawdown(self, value: float) -> None:
        ...

    @property
    def maximum_intra_trade_drawdown(self) -> float:
        """The maximum intra-trade drawdown for all trades (as symbol currency)"""
        ...

    @property.setter
    def maximum_intra_trade_drawdown(self, value: float) -> None:
        ...

    @property
    def profit_loss_standard_deviation(self) -> float:
        """The standard deviation of the profits/losses for all trades (as symbol currency)"""
        ...

    @property.setter
    def profit_loss_standard_deviation(self, value: float) -> None:
        ...

    @property
    def profit_loss_downside_deviation(self) -> float:
        """The downside deviation of the profits/losses for all trades (as symbol currency)"""
        ...

    @property.setter
    def profit_loss_downside_deviation(self, value: float) -> None:
        ...

    @property
    def profit_factor(self) -> float:
        """The ratio of the total profit to the total loss"""
        ...

    @property.setter
    def profit_factor(self, value: float) -> None:
        ...

    @property
    def sharpe_ratio(self) -> float:
        """The ratio of the average profit/loss to the standard deviation"""
        ...

    @property.setter
    def sharpe_ratio(self, value: float) -> None:
        ...

    @property
    def sortino_ratio(self) -> float:
        """The ratio of the average profit/loss to the downside deviation"""
        ...

    @property.setter
    def sortino_ratio(self, value: float) -> None:
        ...

    @property
    def profit_to_max_drawdown_ratio(self) -> float:
        """The ratio of the total profit/loss to the maximum closed trade drawdown"""
        ...

    @property.setter
    def profit_to_max_drawdown_ratio(self, value: float) -> None:
        ...

    @property
    def maximum_end_trade_drawdown(self) -> float:
        """The maximum amount of profit given back by a single trade before exit (as symbol currency)"""
        ...

    @property.setter
    def maximum_end_trade_drawdown(self, value: float) -> None:
        ...

    @property
    def average_end_trade_drawdown(self) -> float:
        """The average amount of profit given back by all trades before exit (as symbol currency)"""
        ...

    @property.setter
    def average_end_trade_drawdown(self, value: float) -> None:
        ...

    @property
    def maximum_drawdown_duration(self) -> datetime.timedelta:
        """The maximum amount of time to recover from a drawdown (longest time between new equity highs or peaks)"""
        ...

    @property.setter
    def maximum_drawdown_duration(self, value: datetime.timedelta) -> None:
        ...

    @property
    def total_fees(self) -> float:
        """The sum of fees for all trades"""
        ...

    @property.setter
    def total_fees(self, value: float) -> None:
        ...

    @overload
    def __init__(self, trades: System.Collections.Generic.IEnumerable[QuantConnect.Statistics.Trade]) -> None:
        """
        Initializes a new instance of the TradeStatistics class
        
        :param trades: The list of closed trades
        """
        ...

    @overload
    def __init__(self) -> None:
        """Initializes a new instance of the TradeStatistics class"""
        ...


class AlgorithmPerformance(System.Object):
    """The AlgorithmPerformance class is a wrapper for TradeStatistics and PortfolioStatistics"""

    @property
    def trade_statistics(self) -> QuantConnect.Statistics.TradeStatistics:
        """The algorithm statistics on closed trades"""
        ...

    @property.setter
    def trade_statistics(self, value: QuantConnect.Statistics.TradeStatistics) -> None:
        ...

    @property
    def portfolio_statistics(self) -> QuantConnect.Statistics.PortfolioStatistics:
        """The algorithm statistics on portfolio"""
        ...

    @property.setter
    def portfolio_statistics(self, value: QuantConnect.Statistics.PortfolioStatistics) -> None:
        ...

    @property
    def closed_trades(self) -> System.Collections.Generic.List[QuantConnect.Statistics.Trade]:
        """The list of closed trades"""
        ...

    @property.setter
    def closed_trades(self, value: System.Collections.Generic.List[QuantConnect.Statistics.Trade]) -> None:
        ...

    @overload
    def __init__(self, trades: System.Collections.Generic.List[QuantConnect.Statistics.Trade], profitLoss: System.Collections.Generic.SortedDictionary[datetime.datetime, float], equity: System.Collections.Generic.SortedDictionary[datetime.datetime, float], portfolioTurnover: System.Collections.Generic.SortedDictionary[datetime.datetime, float], listPerformance: System.Collections.Generic.List[float], listBenchmark: System.Collections.Generic.List[float], startingCapital: float, winningTransactions: int, losingTransactions: int, riskFreeInterestRateModel: QuantConnect.Data.IRiskFreeInterestRateModel, tradingDaysPerYear: int) -> None:
        """
        Initializes a new instance of the AlgorithmPerformance class
        
        :param trades: The list of closed trades
        :param profitLoss: Trade record of profits and losses
        :param equity: The list of daily equity values
        :param portfolioTurnover: The algorithm portfolio turnover
        :param listPerformance: The list of algorithm performance values
        :param listBenchmark: The list of benchmark values
        :param startingCapital: The algorithm starting capital
        :param winningTransactions: Number of winning transactions
        :param losingTransactions: Number of losing transactions
        :param riskFreeInterestRateModel: The risk free interest rate model to use
        :param tradingDaysPerYear: The number of trading days per year
        """
        ...

    @overload
    def __init__(self) -> None:
        """Initializes a new instance of the AlgorithmPerformance class"""
        ...


class StatisticsResults(System.Object):
    """The StatisticsResults class represents total and rolling statistics for an algorithm"""

    @property
    def total_performance(self) -> QuantConnect.Statistics.AlgorithmPerformance:
        """The performance of the algorithm over the whole period"""
        ...

    @property
    def rolling_performances(self) -> System.Collections.Generic.Dictionary[str, QuantConnect.Statistics.AlgorithmPerformance]:
        """The rolling performance of the algorithm over 1, 3, 6, 12 month periods"""
        ...

    @property
    def summary(self) -> System.Collections.Generic.Dictionary[str, str]:
        """Returns a summary of the algorithm performance as a dictionary"""
        ...

    @overload
    def __init__(self, totalPerformance: QuantConnect.Statistics.AlgorithmPerformance, rollingPerformances: System.Collections.Generic.Dictionary[str, QuantConnect.Statistics.AlgorithmPerformance], summary: System.Collections.Generic.Dictionary[str, str]) -> None:
        """
        Initializes a new instance of the StatisticsResults class
        
        :param totalPerformance: The algorithm total performance
        :param rollingPerformances: The algorithm rolling performances
        :param summary: The summary performance dictionary
        """
        ...

    @overload
    def __init__(self) -> None:
        """Initializes a new instance of the StatisticsResults class"""
        ...


class StatisticsBuilder(System.Object):
    """The StatisticsBuilder class creates summary and rolling statistics from trades, equity and benchmark points"""

    @staticmethod
    def create_benchmark_differences(points: System.Collections.Generic.IEnumerable[System.Collections.Generic.KeyValuePair[datetime.datetime, float]], from_date: typing.Union[datetime.datetime, datetime.date], to_date: typing.Union[datetime.datetime, datetime.date]) -> System.Collections.Generic.IEnumerable[System.Collections.Generic.KeyValuePair[datetime.datetime, float]]:
        """
        Yields pairs of date and percentage change for the period
        
        :param points: The values to calculate percentage change for
        :param from_date: Starting date (inclusive)
        :param to_date: Ending date (inclusive)
        :returns: Pairs of date and percentage change.
        """
        ...

    @staticmethod
    def generate(trades: System.Collections.Generic.List[QuantConnect.Statistics.Trade], profit_loss: System.Collections.Generic.SortedDictionary[datetime.datetime, float], points_equity: System.Collections.Generic.List[QuantConnect.ISeriesPoint], points_performance: System.Collections.Generic.List[QuantConnect.ISeriesPoint], points_benchmark: System.Collections.Generic.List[QuantConnect.ISeriesPoint], points_portfolio_turnover: System.Collections.Generic.List[QuantConnect.ISeriesPoint], starting_capital: float, total_fees: float, total_orders: int, estimated_strategy_capacity: QuantConnect.CapacityEstimate, account_currency_symbol: str, transactions: QuantConnect.Securities.SecurityTransactionManager, risk_free_interest_rate_model: QuantConnect.Data.IRiskFreeInterestRateModel, trading_days_per_year: int) -> QuantConnect.Statistics.StatisticsResults:
        """
        Generates the statistics and returns the results
        
        :param trades: The list of closed trades
        :param profit_loss: Trade record of profits and losses
        :param points_equity: The list of daily equity values
        :param points_performance: The list of algorithm performance values
        :param points_benchmark: The list of benchmark values
        :param points_portfolio_turnover: The list of portfolio turnover daily samples
        :param starting_capital: The algorithm starting capital
        :param total_fees: The total fees
        :param total_orders: The total number of transactions
        :param estimated_strategy_capacity: The estimated capacity of this strategy
        :param account_currency_symbol: The account currency symbol
        :param transactions: The transaction manager to get number of winning and losing transactions
        :param risk_free_interest_rate_model: The risk free interest rate model to use
        :param trading_days_per_year: The number of trading days per year
        :returns: Returns a StatisticsResults object.
        """
        ...

    @staticmethod
    def preprocess_performance_values(points: System.Collections.Generic.IEnumerable[System.Collections.Generic.KeyValuePair[datetime.datetime, float]]) -> System.Collections.Generic.IEnumerable[System.Collections.Generic.KeyValuePair[datetime.datetime, float]]:
        """
        Skips the first two entries from the given points and divides each entry by 100
        
        :param points: The values to divide by 100
        :returns: Pairs of date and performance value divided by 100.
        """
        ...


class PerformanceMetrics(System.Object):
    """PerformanceMetrics contains the names of the various performance metrics used for evaluation purposes."""

    ALPHA: str = "Alpha"
    """Algorithm "Alpha" statistic - abnormal returns over the risk free rate and the relationshio (beta) with the benchmark returns."""

    ANNUAL_STANDARD_DEVIATION: str = "Annual Standard Deviation"
    """Annualized standard deviation"""

    ANNUAL_VARIANCE: str = "Annual Variance"
    """Annualized variance statistic calculation using the daily performance variance and trading days per year."""

    AVERAGE_LOSS: str = "Average Loss"
    """The average rate of return for losing trades"""

    AVERAGE_WIN: str = "Average Win"
    """The average rate of return for winning trades"""

    BETA: str = "Beta"
    """Algorithm "beta" statistic - the covariance between the algorithm and benchmark performance, divided by benchmark's variance"""

    COMPOUNDING_ANNUAL_RETURN: str = "Compounding Annual Return"
    """Annual compounded returns statistic based on the final-starting capital and years."""

    DRAWDOWN: str = "Drawdown"
    """Drawdown maximum percentage."""

    ESTIMATED_STRATEGY_CAPACITY: str = "Estimated Strategy Capacity"
    """Total capacity of the algorithm"""

    EXPECTANCY: str = "Expectancy"
    """The expected value of the rate of return"""

    START_EQUITY: str = "Start Equity"
    """Initial Equity Total Value"""

    END_EQUITY: str = "End Equity"
    """Final Equity Total Value"""

    INFORMATION_RATIO: str = "Information Ratio"
    """Information ratio - risk adjusted return"""

    LOSS_RATE: str = "Loss Rate"
    """The ratio of the number of losing trades to the total number of trades"""

    NET_PROFIT: str = "Net Profit"
    """Total net profit percentage"""

    PROBABILISTIC_SHARPE_RATIO: str = "Probabilistic Sharpe Ratio"
    """
    Probabilistic Sharpe Ratio is a probability measure associated with the Sharpe ratio.
    It informs us of the probability that the estimated Sharpe ratio is greater than a chosen benchmark
    """

    PROFIT_LOSS_RATIO: str = "Profit-Loss Ratio"
    """The ratio of the average win rate to the average loss rate"""

    SHARPE_RATIO: str = "Sharpe Ratio"
    """Sharpe ratio with respect to risk free rate: measures excess of return per unit of risk."""

    SORTINO_RATIO: str = "Sortino Ratio"
    """Sortino ratio with respect to risk free rate: measures excess of return per unit of downside risk."""

    TOTAL_FEES: str = "Total Fees"
    """Total amount of fees in the account currency"""

    TOTAL_ORDERS: str = "Total Orders"
    """Total amount of orders in the algorithm"""

    TRACKING_ERROR: str = "Tracking Error"
    """Tracking error volatility (TEV) statistic - a measure of how closely a portfolio follows the index to which it is benchmarked"""

    TREYNOR_RATIO: str = "Treynor Ratio"
    """Treynor ratio statistic is a measurement of the returns earned in excess of that which could have been earned on an investment that has no diversifiable risk"""

    WIN_RATE: str = "Win Rate"
    """The ratio of the number of winning trades to the total number of trades"""

    LOWEST_CAPACITY_ASSET: str = "Lowest Capacity Asset"
    """Provide a reference to the lowest capacity symbol used in scaling down the capacity for debugging."""

    PORTFOLIO_TURNOVER: str = "Portfolio Turnover"
    """The average Portfolio Turnover"""


class FillGroupingMethod(Enum):
    """The method used to group order fills into trades"""

    FILL_TO_FILL = 0
    """A Trade is defined by a fill that establishes or increases a position and an offsetting fill that reduces the position size (0)"""

    FLAT_TO_FLAT = 1
    """A Trade is defined by a sequence of fills, from a flat position to a non-zero position which may increase or decrease in quantity, and back to a flat position (1)"""

    FLAT_TO_REDUCED = 2
    """A Trade is defined by a sequence of fills, from a flat position to a non-zero position and an offsetting fill that reduces the position size (2)"""


class FillMatchingMethod(Enum):
    """The method used to match offsetting order fills"""

    FIFO = 0
    """First In First Out fill matching method (0)"""

    LIFO = 1
    """Last In Last Out fill matching method (1)"""


class Statistics(System.Object):
    """Calculate all the statistics required from the backtest, based on the equity curve and the profit loss statement."""

    @staticmethod
    def annual_downside_standard_deviation(performance: System.Collections.Generic.List[float], trading_days_per_year: float, minimum_acceptable_return: float = 0) -> float:
        """
        Annualized downside standard deviation
        
        :param performance: Collection of double values for daily performance
        :param trading_days_per_year: Number of trading days for the assets in portfolio to get annualize standard deviation.
        :param minimum_acceptable_return: Minimum acceptable return
        :returns: Value for annual downside standard deviation.
        """
        ...

    @staticmethod
    def annual_downside_variance(performance: System.Collections.Generic.List[float], trading_days_per_year: float, minimum_acceptable_return: float = 0) -> float:
        """
        Annualized variance statistic calculation using the daily performance variance and trading days per year.
        
        :param minimum_acceptable_return: Minimum acceptable return
        :returns: Annual variance value.
        """
        ...

    @staticmethod
    def annual_performance(performance: System.Collections.Generic.List[float], trading_days_per_year: float) -> float:
        """
        Annualized return statistic calculated as an average of daily trading performance multiplied by the number of trading days per year.
        
        :param performance: Dictionary collection of double performance values
        :param trading_days_per_year: Trading days per year for the assets in portfolio
        :returns: Double annual performance percentage.
        """
        ...

    @staticmethod
    def annual_standard_deviation(performance: System.Collections.Generic.List[float], trading_days_per_year: float) -> float:
        """
        Annualized standard deviation
        
        :param performance: Collection of double values for daily performance
        :param trading_days_per_year: Number of trading days for the assets in portfolio to get annualize standard deviation.
        :returns: Value for annual standard deviation.
        """
        ...

    @staticmethod
    def annual_variance(performance: System.Collections.Generic.List[float], trading_days_per_year: float) -> float:
        """
        Annualized variance statistic calculation using the daily performance variance and trading days per year.
        
        :returns: Annual variance value.
        """
        ...

    @staticmethod
    def compounding_annual_performance(starting_capital: float, final_capital: float, years: float) -> float:
        """
        Annual compounded returns statistic based on the final-starting capital and years.
        
        :param starting_capital: Algorithm starting capital
        :param final_capital: Algorithm final capital
        :param years: Years trading
        :returns: Decimal fraction for annual compounding performance.
        """
        ...

    @staticmethod
    @overload
    def drawdown_percent(equity_over_time: System.Collections.Generic.SortedDictionary[datetime.datetime, float], rounding: int = 2) -> float:
        """Drawdown maximum percentage."""
        ...

    @staticmethod
    @overload
    def drawdown_percent(current: float, high: float, rounding_decimals: int = 2) -> float:
        """
        Calculate the drawdown between a high and current value
        
        :param current: Current value
        :param high: Latest maximum
        :param rounding_decimals: Digits to round the result too
        :returns: Drawdown percentage.
        """
        ...

    @staticmethod
    def observed_sharpe_ratio(list_performance: System.Collections.Generic.List[float]) -> float:
        """
        Calculates the observed sharpe ratio
        
        :param list_performance: The performance samples to use
        :returns: The observed sharpe ratio.
        """
        ...

    @staticmethod
    def probabilistic_sharpe_ratio(list_performance: System.Collections.Generic.List[float], benchmark_sharpe_ratio: float) -> float:
        """
        Helper method to calculate the probabilistic sharpe ratio
        
        :param list_performance: The list of algorithm performance values
        :param benchmark_sharpe_ratio: The benchmark sharpe ratio to use
        :returns: Probabilistic Sharpe Ratio.
        """
        ...

    @staticmethod
    @overload
    def sharpe_ratio(average_performance: float, standard_deviation: float, risk_free_rate: float) -> float:
        """
        Sharpe ratio with respect to risk free rate: measures excess of return per unit of risk.
        
        :param average_performance: Average daily performance
        :param standard_deviation: Standard deviation of the daily performance
        :param risk_free_rate: The risk free rate
        :returns: Value for sharpe ratio.
        """
        ...

    @staticmethod
    @overload
    def sharpe_ratio(average_performance: float, standard_deviation: float, risk_free_rate: float) -> float:
        """
        Sharpe ratio with respect to risk free rate: measures excess of return per unit of risk.
        
        :param average_performance: Average daily performance
        :param standard_deviation: Standard deviation of the daily performance
        :param risk_free_rate: The risk free rate
        :returns: Value for sharpe ratio.
        """
        ...

    @staticmethod
    @overload
    def sharpe_ratio(algo_performance: System.Collections.Generic.List[float], risk_free_rate: float, trading_days_per_year: float) -> float:
        """
        Sharpe ratio with respect to risk free rate: measures excess of return per unit of risk.
        
        :param algo_performance: Collection of double values for the algorithm daily performance
        :param risk_free_rate: The risk free rate
        :param trading_days_per_year: Trading days per year for the assets in portfolio
        :returns: Value for sharpe ratio.
        """
        ...

    @staticmethod
    def sortino_ratio(algo_performance: System.Collections.Generic.List[float], risk_free_rate: float, trading_days_per_year: float, minimum_acceptable_return: float = 0) -> float:
        """
        Sortino ratio with respect to risk free rate: measures excess of return per unit of downside risk.
        
        :param algo_performance: Collection of double values for the algorithm daily performance
        :param risk_free_rate: The risk free rate
        :param trading_days_per_year: Trading days per year for the assets in portfolio
        :param minimum_acceptable_return: Minimum acceptable return for Sortino ratio calculation
        :returns: Value for Sortino ratio.
        """
        ...

    @staticmethod
    def tracking_error(algo_performance: System.Collections.Generic.List[float], benchmark_performance: System.Collections.Generic.List[float], trading_days_per_year: float) -> float:
        """
        Tracking error volatility (TEV) statistic - a measure of how closely a portfolio follows the index to which it is benchmarked
        
        :param algo_performance: Double collection of algorithm daily performance values
        :param benchmark_performance: Double collection of benchmark daily performance values
        :param trading_days_per_year: Number of trading days per year
        :returns: Value for tracking error.
        """
        ...


class IStatisticsService(metaclass=abc.ABCMeta):
    """This interface exposes methods for accessing algorithm statistics results at runtime."""

    def set_summary_statistic(self, name: str, value: str) -> None:
        """
        Sets or updates a custom summary statistic
        
        :param name: The statistic name
        :param value: The statistic value
        """
        ...

    def statistics_results(self) -> QuantConnect.Statistics.StatisticsResults:
        """
        Calculates and gets the current statistics for the algorithm
        
        :returns: The current statistics.
        """
        ...


class TradeBuilder(System.Object, QuantConnect.Interfaces.ITradeBuilder):
    """The TradeBuilder class generates trades from executions and market price updates"""

    @property
    def closed_trades(self) -> System.Collections.Generic.List[QuantConnect.Statistics.Trade]:
        """The list of closed trades"""
        ...

    def __init__(self, groupingMethod: QuantConnect.Statistics.FillGroupingMethod, matchingMethod: QuantConnect.Statistics.FillMatchingMethod) -> None:
        """Initializes a new instance of the TradeBuilder class"""
        ...

    def apply_split(self, split: QuantConnect.Data.Market.Split, live_mode: bool, data_normalization_mode: QuantConnect.DataNormalizationMode) -> None:
        """
        Applies a split to the trade builder
        
        :param split: The split to be applied
        :param live_mode: True if live mode, false for backtest
        :param data_normalization_mode: The DataNormalizationMode for this security
        """
        ...

    def has_open_position(self, symbol: typing.Union[QuantConnect.Symbol, str]) -> bool:
        """
        Returns true if there is an open position for the symbol
        
        :param symbol: The symbol
        :returns: true if there is an open position for the symbol.
        """
        ...

    def process_fill(self, fill: QuantConnect.Orders.OrderEvent, security_conversion_rate: float, fee_in_account_currency: float, multiplier: float = 1.0) -> None:
        """
        Processes a new fill, eventually creating new trades
        
        :param fill: The new fill order event
        :param security_conversion_rate: The current security market conversion rate into the account currency
        :param fee_in_account_currency: The current order fee in the account currency
        :param multiplier: The contract multiplier
        """
        ...

    def set_live_mode(self, live: bool) -> None:
        """
        Sets the live mode flag
        
        :param live: The live mode flag
        """
        ...

    def set_market_price(self, symbol: typing.Union[QuantConnect.Symbol, str], price: float) -> None:
        """Sets the current market price for the symbol"""
        ...

    def set_security_manager(self, securities: QuantConnect.Securities.SecurityManager) -> None:
        """
        Sets the security manager instance
        
        :param securities: The security manager
        """
        ...





from typing import overload
from enum import Enum
import abc
import typing
import warnings

import System
import System.Collections.Generic
import System.Diagnostics.SymbolStore
import System.Globalization
import System.IO
import System.Reflection
import System.Reflection.Emit
import System.Runtime.InteropServices

System_Reflection_Emit_Label = typing.Any
System_Reflection_Emit_OpCode = typing.Any


class ParameterBuilder(System.Object, metaclass=abc.ABCMeta):
    """This class has no documentation."""

    @property
    def attributes(self) -> int:
        ...

    @property
    def is_in(self) -> bool:
        ...

    @property
    def is_optional(self) -> bool:
        ...

    @property
    def is_out(self) -> bool:
        ...

    @property
    def name(self) -> str:
        ...

    @property
    def position(self) -> int:
        ...

    def __init__(self) -> None:
        """This method is protected."""
        ...

    def set_constant(self, default_value: typing.Any) -> None:
        ...

    @overload
    def set_custom_attribute(self, con: System.Reflection.ConstructorInfo, binary_attribute: typing.List[int]) -> None:
        ...

    @overload
    def set_custom_attribute(self, custom_builder: System.Reflection.Emit.CustomAttributeBuilder) -> None:
        ...

    def set_custom_attribute_core(self, con: System.Reflection.ConstructorInfo, binary_attribute: System.ReadOnlySpan[int]) -> None:
        """This method is protected."""
        ...


class Label(System.IEquatable[System_Reflection_Emit_Label]):
    """Represents a label in the instruction stream. Used in conjunction with the ILGenerator class."""

    @property
    def id(self) -> int:
        """Gets the label unique id assigned by the ILGenerator."""
        ...

    @overload
    def equals(self, obj: typing.Any) -> bool:
        ...

    @overload
    def equals(self, obj: System.Reflection.Emit.Label) -> bool:
        ...

    def get_hash_code(self) -> int:
        ...


class OperandType(Enum):
    """Describes the operand type of IL instruction."""

    INLINE_BR_TARGET = 0

    INLINE_FIELD = 1

    INLINE_I = 2

    INLINE_I_8 = 3

    INLINE_METHOD = 4

    INLINE_NONE = 5

    INLINE_PHI = 6
    """OperandType.InlinePhi has been deprecated and is not supported."""

    INLINE_R = 7

    INLINE_SIG = 9

    INLINE_STRING = 10

    INLINE_SWITCH = 11

    INLINE_TOK = 12

    INLINE_TYPE = 13

    INLINE_VAR = 14

    SHORT_INLINE_BR_TARGET = 15

    SHORT_INLINE_I = 16

    SHORT_INLINE_R = 17

    SHORT_INLINE_VAR = 18


class FlowControl(Enum):
    """Describes how an instruction alters the flow of control."""

    BRANCH = 0

    BREAK = 1

    CALL = 2

    COND_BRANCH = 3

    META = 4

    NEXT = 5

    PHI = 6
    """FlowControl.Phi has been deprecated and is not supported."""

    RETURN = 7

    THROW = 8


class OpCodeType(Enum):
    """Describes the types of the IL instructions."""

    ANNOTATION = 0
    """OpCodeType.Annotation has been deprecated and is not supported."""

    MACRO = 1

    NTERNAL = 2

    OBJMODEL = 3

    PREFIX = 4

    PRIMITIVE = 5


class StackBehaviour(Enum):
    """Describes how values are pushed onto a stack or popped off a stack."""

    POP_0 = 0

    POP_1 = 1

    POP_1_POP_1 = 2

    POPI = 3

    POPI_POP_1 = 4

    POPI_POPI = 5

    POPI_POPI_8 = 6

    POPI_POPI_POPI = 7

    POPI_POPR_4 = 8

    POPI_POPR_8 = 9

    POPREF = 10

    POPREF_POP_1 = 11

    POPREF_POPI = 12

    POPREF_POPI_POPI = 13

    POPREF_POPI_POPI_8 = 14

    POPREF_POPI_POPR_4 = 15

    POPREF_POPI_POPR_8 = 16

    POPREF_POPI_POPREF = 17

    PUSH_0 = 18

    PUSH_1 = 19

    PUSH_1_PUSH_1 = 20

    PUSHI = 21

    PUSHI_8 = 22

    PUSHR_4 = 23

    PUSHR_8 = 24

    PUSHREF = 25

    VARPOP = 26

    VARPUSH = 27

    POPREF_POPI_POP_1 = 28


class OpCode(System.IEquatable[System_Reflection_Emit_OpCode]):
    """This class has no documentation."""

    @property
    def evaluation_stack_delta(self) -> int:
        """The value of how the IL instruction changes the evaluation stack."""
        ...

    @property
    def operand_type(self) -> System.Reflection.Emit.OperandType:
        ...

    @property
    def flow_control(self) -> System.Reflection.Emit.FlowControl:
        ...

    @property
    def op_code_type(self) -> System.Reflection.Emit.OpCodeType:
        ...

    @property
    def stack_behaviour_pop(self) -> System.Reflection.Emit.StackBehaviour:
        ...

    @property
    def stack_behaviour_push(self) -> System.Reflection.Emit.StackBehaviour:
        ...

    @property
    def size(self) -> int:
        ...

    @property
    def value(self) -> int:
        ...

    @property
    def name(self) -> str:
        ...

    @overload
    def equals(self, obj: typing.Any) -> bool:
        ...

    @overload
    def equals(self, obj: System.Reflection.Emit.OpCode) -> bool:
        ...

    def get_hash_code(self) -> int:
        ...

    def to_string(self) -> str:
        ...


class LocalBuilder(System.Reflection.LocalVariableInfo, metaclass=abc.ABCMeta):
    """This class has no documentation."""

    def __init__(self) -> None:
        """
        Initializes a new instance of the LocalBuilder class.
        
        This method is protected.
        """
        ...

    def set_local_sym_info(self, name: str) -> None:
        """
        Sets the name of this local variable.
        
        :param name: The name of the local variable
        """
        ...

    def set_local_sym_info_core(self, name: str) -> None:
        """
        When overridden in a derived class, sets the name of this local variable.
        
        This method is protected.
        
        :param name: The name of the local variable.
        """
        ...


class ILGenerator(System.Object, metaclass=abc.ABCMeta):
    """This class has no documentation."""

    @property
    @abc.abstractmethod
    def il_offset(self) -> int:
        ...

    def __init__(self) -> None:
        """This method is protected."""
        ...

    def begin_catch_block(self, exception_type: typing.Type) -> None:
        ...

    def begin_except_filter_block(self) -> None:
        ...

    def begin_exception_block(self) -> System.Reflection.Emit.Label:
        ...

    def begin_fault_block(self) -> None:
        ...

    def begin_finally_block(self) -> None:
        ...

    def begin_scope(self) -> None:
        ...

    @staticmethod
    def create_label(id: int) -> System.Reflection.Emit.Label:
        """
        Creates a Label with the given id.
        
        This method is protected.
        
        :param id: The unique id for the label.
        :returns: The Label created.
        """
        ...

    @overload
    def declare_local(self, local_type: typing.Type) -> System.Reflection.Emit.LocalBuilder:
        ...

    @overload
    def declare_local(self, local_type: typing.Type, pinned: bool) -> System.Reflection.Emit.LocalBuilder:
        ...

    def define_label(self) -> System.Reflection.Emit.Label:
        ...

    @overload
    def emit(self, opcode: System.Reflection.Emit.OpCode) -> None:
        ...

    @overload
    def emit(self, opcode: System.Reflection.Emit.OpCode, arg: int) -> None:
        ...

    @overload
    def emit(self, opcode: System.Reflection.Emit.OpCode, arg: int) -> None:
        ...

    @overload
    def emit(self, opcode: System.Reflection.Emit.OpCode, arg: int) -> None:
        ...

    @overload
    def emit(self, opcode: System.Reflection.Emit.OpCode, arg: float) -> None:
        ...

    @overload
    def emit(self, opcode: System.Reflection.Emit.OpCode, arg: float) -> None:
        ...

    @overload
    def emit(self, opcode: System.Reflection.Emit.OpCode, arg: int) -> None:
        ...

    @overload
    def emit(self, opcode: System.Reflection.Emit.OpCode, meth: System.Reflection.MethodInfo) -> None:
        ...

    @overload
    def emit(self, opcode: System.Reflection.Emit.OpCode, signature: System.Reflection.Emit.SignatureHelper) -> None:
        ...

    @overload
    def emit(self, opcode: System.Reflection.Emit.OpCode, con: System.Reflection.ConstructorInfo) -> None:
        ...

    @overload
    def emit(self, opcode: System.Reflection.Emit.OpCode, cls: typing.Type) -> None:
        ...

    @overload
    def emit(self, opcode: System.Reflection.Emit.OpCode, label: System.Reflection.Emit.Label) -> None:
        ...

    @overload
    def emit(self, opcode: System.Reflection.Emit.OpCode, labels: typing.List[System.Reflection.Emit.Label]) -> None:
        ...

    @overload
    def emit(self, opcode: System.Reflection.Emit.OpCode, field: System.Reflection.FieldInfo) -> None:
        ...

    @overload
    def emit(self, opcode: System.Reflection.Emit.OpCode, str: str) -> None:
        ...

    @overload
    def emit(self, opcode: System.Reflection.Emit.OpCode, local: System.Reflection.Emit.LocalBuilder) -> None:
        ...

    @overload
    def emit(self, opcode: System.Reflection.Emit.OpCode, arg: int) -> None:
        ...

    def emit_call(self, opcode: System.Reflection.Emit.OpCode, method_info: System.Reflection.MethodInfo, optional_parameter_types: typing.List[typing.Type]) -> None:
        ...

    @overload
    def emit_calli(self, opcode: System.Reflection.Emit.OpCode, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, parameter_types: typing.List[typing.Type], optional_parameter_types: typing.List[typing.Type]) -> None:
        ...

    @overload
    def emit_calli(self, opcode: System.Reflection.Emit.OpCode, unmanaged_call_conv: System.Runtime.InteropServices.CallingConvention, return_type: typing.Type, parameter_types: typing.List[typing.Type]) -> None:
        ...

    @overload
    def emit_write_line(self, value: str) -> None:
        ...

    @overload
    def emit_write_line(self, local_builder: System.Reflection.Emit.LocalBuilder) -> None:
        ...

    @overload
    def emit_write_line(self, fld: System.Reflection.FieldInfo) -> None:
        ...

    def end_exception_block(self) -> None:
        ...

    def end_scope(self) -> None:
        ...

    def mark_label(self, loc: System.Reflection.Emit.Label) -> None:
        ...

    def mark_sequence_point(self, document: System.Diagnostics.SymbolStore.ISymbolDocumentWriter, start_line: int, start_column: int, end_line: int, end_column: int) -> None:
        """
        Marks a sequence point in the Microsoft intermediate language (MSIL) stream.
        
        :param document: The document for which the sequence point is being defined.
        :param start_line: The line where the sequence point begins.
        :param start_column: The column in the line where the sequence point begins.
        :param end_line: The line where the sequence point ends.
        :param end_column: The column in the line where the sequence point ends.
        """
        ...

    def mark_sequence_point_core(self, document: System.Diagnostics.SymbolStore.ISymbolDocumentWriter, start_line: int, start_column: int, end_line: int, end_column: int) -> None:
        """
        When overridden in a derived class, marks a sequence point in the Microsoft intermediate language (MSIL) stream.
        
        This method is protected.
        
        :param document: The document for which the sequence point is being defined.
        :param start_line: The line where the sequence point begins.
        :param start_column: The column in the line where the sequence point begins.
        :param end_line: The line where the sequence point ends.
        :param end_column: The column in the line where the sequence point ends.
        """
        ...

    def throw_exception(self, exc_type: typing.Type) -> None:
        ...

    def using_namespace(self, using_namespace: str) -> None:
        ...


class ConstructorBuilder(System.Reflection.ConstructorInfo, metaclass=abc.ABCMeta):
    """This class has no documentation."""

    @property
    def init_locals(self) -> bool:
        ...

    @property.setter
    def init_locals(self, value: bool) -> None:
        ...

    @property
    @abc.abstractmethod
    def init_locals_core(self) -> bool:
        """This property is protected."""
        ...

    @property.setter
    def init_locals_core(self, value: bool) -> None:
        ...

    def __init__(self) -> None:
        """This method is protected."""
        ...

    def define_parameter(self, i_sequence: int, attributes: System.Reflection.ParameterAttributes, str_param_name: str) -> System.Reflection.Emit.ParameterBuilder:
        ...

    def define_parameter_core(self, i_sequence: int, attributes: System.Reflection.ParameterAttributes, str_param_name: str) -> System.Reflection.Emit.ParameterBuilder:
        """This method is protected."""
        ...

    @overload
    def get_il_generator(self) -> System.Reflection.Emit.ILGenerator:
        ...

    @overload
    def get_il_generator(self, stream_size: int) -> System.Reflection.Emit.ILGenerator:
        ...

    def get_il_generator_core(self, stream_size: int) -> System.Reflection.Emit.ILGenerator:
        """This method is protected."""
        ...

    @overload
    def set_custom_attribute(self, con: System.Reflection.ConstructorInfo, binary_attribute: typing.List[int]) -> None:
        ...

    @overload
    def set_custom_attribute(self, custom_builder: System.Reflection.Emit.CustomAttributeBuilder) -> None:
        ...

    def set_custom_attribute_core(self, con: System.Reflection.ConstructorInfo, binary_attribute: System.ReadOnlySpan[int]) -> None:
        """This method is protected."""
        ...

    def set_implementation_flags(self, attributes: System.Reflection.MethodImplAttributes) -> None:
        ...

    def set_implementation_flags_core(self, attributes: System.Reflection.MethodImplAttributes) -> None:
        """This method is protected."""
        ...


class OpCodes(System.Object):
    """
    The IL instruction opcodes supported by the
          runtime. The IL Instruction Specification describes each
          Opcode.
    """

    NOP: System.Reflection.Emit.OpCode = ...

    BREAK: System.Reflection.Emit.OpCode = ...

    LDARG_0: System.Reflection.Emit.OpCode = ...

    LDARG_1: System.Reflection.Emit.OpCode = ...

    LDARG_2: System.Reflection.Emit.OpCode = ...

    LDARG_3: System.Reflection.Emit.OpCode = ...

    LDLOC_0: System.Reflection.Emit.OpCode = ...

    LDLOC_1: System.Reflection.Emit.OpCode = ...

    LDLOC_2: System.Reflection.Emit.OpCode = ...

    LDLOC_3: System.Reflection.Emit.OpCode = ...

    STLOC_0: System.Reflection.Emit.OpCode = ...

    STLOC_1: System.Reflection.Emit.OpCode = ...

    STLOC_2: System.Reflection.Emit.OpCode = ...

    STLOC_3: System.Reflection.Emit.OpCode = ...

    LDARG_S: System.Reflection.Emit.OpCode = ...

    LDARGA_S: System.Reflection.Emit.OpCode = ...

    STARG_S: System.Reflection.Emit.OpCode = ...

    LDLOC_S: System.Reflection.Emit.OpCode = ...

    LDLOCA_S: System.Reflection.Emit.OpCode = ...

    STLOC_S: System.Reflection.Emit.OpCode = ...

    LDNULL: System.Reflection.Emit.OpCode = ...

    LDC_I_4_M_1: System.Reflection.Emit.OpCode = ...

    LDC_I_4_0: System.Reflection.Emit.OpCode = ...

    LDC_I_4_1: System.Reflection.Emit.OpCode = ...

    LDC_I_4_2: System.Reflection.Emit.OpCode = ...

    LDC_I_4_3: System.Reflection.Emit.OpCode = ...

    LDC_I_4_4: System.Reflection.Emit.OpCode = ...

    LDC_I_4_5: System.Reflection.Emit.OpCode = ...

    LDC_I_4_6: System.Reflection.Emit.OpCode = ...

    LDC_I_4_7: System.Reflection.Emit.OpCode = ...

    LDC_I_4_8: System.Reflection.Emit.OpCode = ...

    LDC_I_4_S: System.Reflection.Emit.OpCode = ...

    LDC_I_4: System.Reflection.Emit.OpCode = ...

    LDC_I_8: System.Reflection.Emit.OpCode = ...

    LDC_R_4: System.Reflection.Emit.OpCode = ...

    LDC_R_8: System.Reflection.Emit.OpCode = ...

    DUP: System.Reflection.Emit.OpCode = ...

    POP: System.Reflection.Emit.OpCode = ...

    JMP: System.Reflection.Emit.OpCode = ...

    CALL: System.Reflection.Emit.OpCode = ...

    CALLI: System.Reflection.Emit.OpCode = ...

    RET: System.Reflection.Emit.OpCode = ...

    BR_S: System.Reflection.Emit.OpCode = ...

    BRFALSE_S: System.Reflection.Emit.OpCode = ...

    BRTRUE_S: System.Reflection.Emit.OpCode = ...

    BEQ_S: System.Reflection.Emit.OpCode = ...

    BGE_S: System.Reflection.Emit.OpCode = ...

    BGT_S: System.Reflection.Emit.OpCode = ...

    BLE_S: System.Reflection.Emit.OpCode = ...

    BLT_S: System.Reflection.Emit.OpCode = ...

    BNE_UN_S: System.Reflection.Emit.OpCode = ...

    BGE_UN_S: System.Reflection.Emit.OpCode = ...

    BGT_UN_S: System.Reflection.Emit.OpCode = ...

    BLE_UN_S: System.Reflection.Emit.OpCode = ...

    BLT_UN_S: System.Reflection.Emit.OpCode = ...

    BR: System.Reflection.Emit.OpCode = ...

    BRFALSE: System.Reflection.Emit.OpCode = ...

    BRTRUE: System.Reflection.Emit.OpCode = ...

    BEQ: System.Reflection.Emit.OpCode = ...

    BGE: System.Reflection.Emit.OpCode = ...

    BGT: System.Reflection.Emit.OpCode = ...

    BLE: System.Reflection.Emit.OpCode = ...

    BLT: System.Reflection.Emit.OpCode = ...

    BNE_UN: System.Reflection.Emit.OpCode = ...

    BGE_UN: System.Reflection.Emit.OpCode = ...

    BGT_UN: System.Reflection.Emit.OpCode = ...

    BLE_UN: System.Reflection.Emit.OpCode = ...

    BLT_UN: System.Reflection.Emit.OpCode = ...

    SWITCH: System.Reflection.Emit.OpCode = ...

    LDIND_I_1: System.Reflection.Emit.OpCode = ...

    LDIND_U_1: System.Reflection.Emit.OpCode = ...

    LDIND_I_2: System.Reflection.Emit.OpCode = ...

    LDIND_U_2: System.Reflection.Emit.OpCode = ...

    LDIND_I_4: System.Reflection.Emit.OpCode = ...

    LDIND_U_4: System.Reflection.Emit.OpCode = ...

    LDIND_I_8: System.Reflection.Emit.OpCode = ...

    LDIND_I: System.Reflection.Emit.OpCode = ...

    LDIND_R_4: System.Reflection.Emit.OpCode = ...

    LDIND_R_8: System.Reflection.Emit.OpCode = ...

    LDIND_REF: System.Reflection.Emit.OpCode = ...

    STIND_REF: System.Reflection.Emit.OpCode = ...

    STIND_I_1: System.Reflection.Emit.OpCode = ...

    STIND_I_2: System.Reflection.Emit.OpCode = ...

    STIND_I_4: System.Reflection.Emit.OpCode = ...

    STIND_I_8: System.Reflection.Emit.OpCode = ...

    STIND_R_4: System.Reflection.Emit.OpCode = ...

    STIND_R_8: System.Reflection.Emit.OpCode = ...

    ADD: System.Reflection.Emit.OpCode = ...

    SUB: System.Reflection.Emit.OpCode = ...

    MUL: System.Reflection.Emit.OpCode = ...

    DIV: System.Reflection.Emit.OpCode = ...

    DIV_UN: System.Reflection.Emit.OpCode = ...

    REM: System.Reflection.Emit.OpCode = ...

    REM_UN: System.Reflection.Emit.OpCode = ...

    AND: System.Reflection.Emit.OpCode = ...

    OR: System.Reflection.Emit.OpCode = ...

    XOR: System.Reflection.Emit.OpCode = ...

    SHL: System.Reflection.Emit.OpCode = ...

    SHR: System.Reflection.Emit.OpCode = ...

    SHR_UN: System.Reflection.Emit.OpCode = ...

    NEG: System.Reflection.Emit.OpCode = ...

    NOT: System.Reflection.Emit.OpCode = ...

    CONV_I_1: System.Reflection.Emit.OpCode = ...

    CONV_I_2: System.Reflection.Emit.OpCode = ...

    CONV_I_4: System.Reflection.Emit.OpCode = ...

    CONV_I_8: System.Reflection.Emit.OpCode = ...

    CONV_R_4: System.Reflection.Emit.OpCode = ...

    CONV_R_8: System.Reflection.Emit.OpCode = ...

    CONV_U_4: System.Reflection.Emit.OpCode = ...

    CONV_U_8: System.Reflection.Emit.OpCode = ...

    CALLVIRT: System.Reflection.Emit.OpCode = ...

    CPOBJ: System.Reflection.Emit.OpCode = ...

    LDOBJ: System.Reflection.Emit.OpCode = ...

    LDSTR: System.Reflection.Emit.OpCode = ...

    NEWOBJ: System.Reflection.Emit.OpCode = ...

    CASTCLASS: System.Reflection.Emit.OpCode = ...

    ISINST: System.Reflection.Emit.OpCode = ...

    CONV_R_UN: System.Reflection.Emit.OpCode = ...

    UNBOX: System.Reflection.Emit.OpCode = ...

    THROW: System.Reflection.Emit.OpCode = ...

    LDFLD: System.Reflection.Emit.OpCode = ...

    LDFLDA: System.Reflection.Emit.OpCode = ...

    STFLD: System.Reflection.Emit.OpCode = ...

    LDSFLD: System.Reflection.Emit.OpCode = ...

    LDSFLDA: System.Reflection.Emit.OpCode = ...

    STSFLD: System.Reflection.Emit.OpCode = ...

    STOBJ: System.Reflection.Emit.OpCode = ...

    CONV_OVF_I_1_UN: System.Reflection.Emit.OpCode = ...

    CONV_OVF_I_2_UN: System.Reflection.Emit.OpCode = ...

    CONV_OVF_I_4_UN: System.Reflection.Emit.OpCode = ...

    CONV_OVF_I_8_UN: System.Reflection.Emit.OpCode = ...

    CONV_OVF_U_1_UN: System.Reflection.Emit.OpCode = ...

    CONV_OVF_U_2_UN: System.Reflection.Emit.OpCode = ...

    CONV_OVF_U_4_UN: System.Reflection.Emit.OpCode = ...

    CONV_OVF_U_8_UN: System.Reflection.Emit.OpCode = ...

    CONV_OVF_I_UN: System.Reflection.Emit.OpCode = ...

    CONV_OVF_U_UN: System.Reflection.Emit.OpCode = ...

    BOX: System.Reflection.Emit.OpCode = ...

    NEWARR: System.Reflection.Emit.OpCode = ...

    LDLEN: System.Reflection.Emit.OpCode = ...

    LDELEMA: System.Reflection.Emit.OpCode = ...

    LDELEM_I_1: System.Reflection.Emit.OpCode = ...

    LDELEM_U_1: System.Reflection.Emit.OpCode = ...

    LDELEM_I_2: System.Reflection.Emit.OpCode = ...

    LDELEM_U_2: System.Reflection.Emit.OpCode = ...

    LDELEM_I_4: System.Reflection.Emit.OpCode = ...

    LDELEM_U_4: System.Reflection.Emit.OpCode = ...

    LDELEM_I_8: System.Reflection.Emit.OpCode = ...

    LDELEM_I: System.Reflection.Emit.OpCode = ...

    LDELEM_R_4: System.Reflection.Emit.OpCode = ...

    LDELEM_R_8: System.Reflection.Emit.OpCode = ...

    LDELEM_REF: System.Reflection.Emit.OpCode = ...

    STELEM_I: System.Reflection.Emit.OpCode = ...

    STELEM_I_1: System.Reflection.Emit.OpCode = ...

    STELEM_I_2: System.Reflection.Emit.OpCode = ...

    STELEM_I_4: System.Reflection.Emit.OpCode = ...

    STELEM_I_8: System.Reflection.Emit.OpCode = ...

    STELEM_R_4: System.Reflection.Emit.OpCode = ...

    STELEM_R_8: System.Reflection.Emit.OpCode = ...

    STELEM_REF: System.Reflection.Emit.OpCode = ...

    LDELEM: System.Reflection.Emit.OpCode = ...

    STELEM: System.Reflection.Emit.OpCode = ...

    UNBOX_ANY: System.Reflection.Emit.OpCode = ...

    CONV_OVF_I_1: System.Reflection.Emit.OpCode = ...

    CONV_OVF_U_1: System.Reflection.Emit.OpCode = ...

    CONV_OVF_I_2: System.Reflection.Emit.OpCode = ...

    CONV_OVF_U_2: System.Reflection.Emit.OpCode = ...

    CONV_OVF_I_4: System.Reflection.Emit.OpCode = ...

    CONV_OVF_U_4: System.Reflection.Emit.OpCode = ...

    CONV_OVF_I_8: System.Reflection.Emit.OpCode = ...

    CONV_OVF_U_8: System.Reflection.Emit.OpCode = ...

    REFANYVAL: System.Reflection.Emit.OpCode = ...

    CKFINITE: System.Reflection.Emit.OpCode = ...

    MKREFANY: System.Reflection.Emit.OpCode = ...

    LDTOKEN: System.Reflection.Emit.OpCode = ...

    CONV_U_2: System.Reflection.Emit.OpCode = ...

    CONV_U_1: System.Reflection.Emit.OpCode = ...

    CONV_I: System.Reflection.Emit.OpCode = ...

    CONV_OVF_I: System.Reflection.Emit.OpCode = ...

    CONV_OVF_U: System.Reflection.Emit.OpCode = ...

    ADD_OVF: System.Reflection.Emit.OpCode = ...

    ADD_OVF_UN: System.Reflection.Emit.OpCode = ...

    MUL_OVF: System.Reflection.Emit.OpCode = ...

    MUL_OVF_UN: System.Reflection.Emit.OpCode = ...

    SUB_OVF: System.Reflection.Emit.OpCode = ...

    SUB_OVF_UN: System.Reflection.Emit.OpCode = ...

    ENDFINALLY: System.Reflection.Emit.OpCode = ...

    LEAVE: System.Reflection.Emit.OpCode = ...

    LEAVE_S: System.Reflection.Emit.OpCode = ...

    STIND_I: System.Reflection.Emit.OpCode = ...

    CONV_U: System.Reflection.Emit.OpCode = ...

    PREFIX_7: System.Reflection.Emit.OpCode = ...

    PREFIX_6: System.Reflection.Emit.OpCode = ...

    PREFIX_5: System.Reflection.Emit.OpCode = ...

    PREFIX_4: System.Reflection.Emit.OpCode = ...

    PREFIX_3: System.Reflection.Emit.OpCode = ...

    PREFIX_2: System.Reflection.Emit.OpCode = ...

    PREFIX_1: System.Reflection.Emit.OpCode = ...

    PREFIXREF: System.Reflection.Emit.OpCode = ...

    ARGLIST: System.Reflection.Emit.OpCode = ...

    CEQ: System.Reflection.Emit.OpCode = ...

    CGT: System.Reflection.Emit.OpCode = ...

    CGT_UN: System.Reflection.Emit.OpCode = ...

    CLT: System.Reflection.Emit.OpCode = ...

    CLT_UN: System.Reflection.Emit.OpCode = ...

    LDFTN: System.Reflection.Emit.OpCode = ...

    LDVIRTFTN: System.Reflection.Emit.OpCode = ...

    LDARG: System.Reflection.Emit.OpCode = ...

    LDARGA: System.Reflection.Emit.OpCode = ...

    STARG: System.Reflection.Emit.OpCode = ...

    LDLOC: System.Reflection.Emit.OpCode = ...

    LDLOCA: System.Reflection.Emit.OpCode = ...

    STLOC: System.Reflection.Emit.OpCode = ...

    LOCALLOC: System.Reflection.Emit.OpCode = ...

    ENDFILTER: System.Reflection.Emit.OpCode = ...

    UNALIGNED: System.Reflection.Emit.OpCode = ...

    VOLATILE: System.Reflection.Emit.OpCode = ...

    TAILCALL: System.Reflection.Emit.OpCode = ...

    INITOBJ: System.Reflection.Emit.OpCode = ...

    CONSTRAINED: System.Reflection.Emit.OpCode = ...

    CPBLK: System.Reflection.Emit.OpCode = ...

    INITBLK: System.Reflection.Emit.OpCode = ...

    RETHROW: System.Reflection.Emit.OpCode = ...

    SIZEOF: System.Reflection.Emit.OpCode = ...

    REFANYTYPE: System.Reflection.Emit.OpCode = ...

    READONLY: System.Reflection.Emit.OpCode = ...

    @staticmethod
    def takes_single_byte_argument(inst: System.Reflection.Emit.OpCode) -> bool:
        ...


class FieldBuilder(System.Reflection.FieldInfo, metaclass=abc.ABCMeta):
    """This class has no documentation."""

    def __init__(self) -> None:
        """This method is protected."""
        ...

    def set_constant(self, default_value: typing.Any) -> None:
        ...

    def set_constant_core(self, default_value: typing.Any) -> None:
        """This method is protected."""
        ...

    @overload
    def set_custom_attribute(self, con: System.Reflection.ConstructorInfo, binary_attribute: typing.List[int]) -> None:
        ...

    @overload
    def set_custom_attribute(self, custom_builder: System.Reflection.Emit.CustomAttributeBuilder) -> None:
        ...

    def set_custom_attribute_core(self, con: System.Reflection.ConstructorInfo, binary_attribute: System.ReadOnlySpan[int]) -> None:
        """This method is protected."""
        ...

    def set_offset(self, i_offset: int) -> None:
        ...

    def set_offset_core(self, i_offset: int) -> None:
        """This method is protected."""
        ...


class EnumBuilder(System.Reflection.TypeInfo, metaclass=abc.ABCMeta):
    """This class has no documentation."""

    @property
    def underlying_field(self) -> System.Reflection.Emit.FieldBuilder:
        ...

    @property
    @abc.abstractmethod
    def underlying_field_core(self) -> System.Reflection.Emit.FieldBuilder:
        """This property is protected."""
        ...

    def __init__(self) -> None:
        """This method is protected."""
        ...

    def create_type(self) -> typing.Type:
        ...

    def create_type_info(self) -> System.Reflection.TypeInfo:
        ...

    def create_type_info_core(self) -> System.Reflection.TypeInfo:
        """This method is protected."""
        ...

    def define_literal(self, literal_name: str, literal_value: typing.Any) -> System.Reflection.Emit.FieldBuilder:
        ...

    def define_literal_core(self, literal_name: str, literal_value: typing.Any) -> System.Reflection.Emit.FieldBuilder:
        """This method is protected."""
        ...

    @overload
    def make_array_type(self) -> typing.Type:
        ...

    @overload
    def make_array_type(self, rank: int) -> typing.Type:
        ...

    def make_by_ref_type(self) -> typing.Type:
        ...

    def make_pointer_type(self) -> typing.Type:
        ...

    @overload
    def set_custom_attribute(self, con: System.Reflection.ConstructorInfo, binary_attribute: typing.List[int]) -> None:
        ...

    @overload
    def set_custom_attribute(self, custom_builder: System.Reflection.Emit.CustomAttributeBuilder) -> None:
        ...

    def set_custom_attribute_core(self, con: System.Reflection.ConstructorInfo, binary_attribute: System.ReadOnlySpan[int]) -> None:
        """This method is protected."""
        ...


class GenericTypeParameterBuilder(System.Reflection.TypeInfo, metaclass=abc.ABCMeta):
    """This class has no documentation."""

    def __init__(self) -> None:
        """This method is protected."""
        ...

    def set_base_type_constraint(self, base_type_constraint: typing.Type) -> None:
        ...

    def set_base_type_constraint_core(self, base_type_constraint: typing.Type) -> None:
        """This method is protected."""
        ...

    @overload
    def set_custom_attribute(self, con: System.Reflection.ConstructorInfo, binary_attribute: typing.List[int]) -> None:
        ...

    @overload
    def set_custom_attribute(self, custom_builder: System.Reflection.Emit.CustomAttributeBuilder) -> None:
        ...

    def set_custom_attribute_core(self, con: System.Reflection.ConstructorInfo, binary_attribute: System.ReadOnlySpan[int]) -> None:
        """This method is protected."""
        ...

    def set_generic_parameter_attributes(self, generic_parameter_attributes: System.Reflection.GenericParameterAttributes) -> None:
        ...

    def set_generic_parameter_attributes_core(self, generic_parameter_attributes: System.Reflection.GenericParameterAttributes) -> None:
        """This method is protected."""
        ...

    def set_interface_constraints(self, *interface_constraints: typing.Type) -> None:
        ...

    def set_interface_constraints_core(self, *interface_constraints: typing.Type) -> None:
        """This method is protected."""
        ...


class MethodBuilder(System.Reflection.MethodInfo, metaclass=abc.ABCMeta):
    """This class has no documentation."""

    @property
    def init_locals(self) -> bool:
        ...

    @property.setter
    def init_locals(self, value: bool) -> None:
        ...

    @property
    @abc.abstractmethod
    def init_locals_core(self) -> bool:
        """This property is protected."""
        ...

    @property.setter
    def init_locals_core(self, value: bool) -> None:
        ...

    def __init__(self) -> None:
        """This method is protected."""
        ...

    def define_generic_parameters(self, *names: str) -> typing.List[System.Reflection.Emit.GenericTypeParameterBuilder]:
        ...

    def define_generic_parameters_core(self, *names: str) -> typing.List[System.Reflection.Emit.GenericTypeParameterBuilder]:
        """This method is protected."""
        ...

    def define_parameter(self, position: int, attributes: System.Reflection.ParameterAttributes, str_param_name: str) -> System.Reflection.Emit.ParameterBuilder:
        ...

    def define_parameter_core(self, position: int, attributes: System.Reflection.ParameterAttributes, str_param_name: str) -> System.Reflection.Emit.ParameterBuilder:
        """This method is protected."""
        ...

    @overload
    def get_il_generator(self) -> System.Reflection.Emit.ILGenerator:
        ...

    @overload
    def get_il_generator(self, size: int) -> System.Reflection.Emit.ILGenerator:
        ...

    def get_il_generator_core(self, size: int) -> System.Reflection.Emit.ILGenerator:
        """This method is protected."""
        ...

    @overload
    def set_custom_attribute(self, con: System.Reflection.ConstructorInfo, binary_attribute: typing.List[int]) -> None:
        ...

    @overload
    def set_custom_attribute(self, custom_builder: System.Reflection.Emit.CustomAttributeBuilder) -> None:
        ...

    def set_custom_attribute_core(self, con: System.Reflection.ConstructorInfo, binary_attribute: System.ReadOnlySpan[int]) -> None:
        """This method is protected."""
        ...

    def set_implementation_flags(self, attributes: System.Reflection.MethodImplAttributes) -> None:
        ...

    def set_implementation_flags_core(self, attributes: System.Reflection.MethodImplAttributes) -> None:
        """This method is protected."""
        ...

    def set_parameters(self, *parameter_types: typing.Type) -> None:
        ...

    def set_return_type(self, return_type: typing.Type) -> None:
        ...

    def set_signature(self, return_type: typing.Type, return_type_required_custom_modifiers: typing.List[typing.Type], return_type_optional_custom_modifiers: typing.List[typing.Type], parameter_types: typing.List[typing.Type], parameter_type_required_custom_modifiers: typing.List[typing.List[typing.Type]], parameter_type_optional_custom_modifiers: typing.List[typing.List[typing.Type]]) -> None:
        ...

    def set_signature_core(self, return_type: typing.Type, return_type_required_custom_modifiers: typing.List[typing.Type], return_type_optional_custom_modifiers: typing.List[typing.Type], parameter_types: typing.List[typing.Type], parameter_type_required_custom_modifiers: typing.List[typing.List[typing.Type]], parameter_type_optional_custom_modifiers: typing.List[typing.List[typing.Type]]) -> None:
        """This method is protected."""
        ...


class PackingSize(Enum):
    """This class has no documentation."""

    UNSPECIFIED = 0

    SIZE_1 = 1

    SIZE_2 = 2

    SIZE_4 = 4

    SIZE_8 = 8

    SIZE_16 = 16

    SIZE_32 = 32

    SIZE_64 = 64

    SIZE_128 = 128


class EventBuilder(System.Object, metaclass=abc.ABCMeta):
    """This class has no documentation."""

    def __init__(self) -> None:
        """This method is protected."""
        ...

    def add_other_method(self, md_builder: System.Reflection.Emit.MethodBuilder) -> None:
        ...

    def add_other_method_core(self, md_builder: System.Reflection.Emit.MethodBuilder) -> None:
        """This method is protected."""
        ...

    def set_add_on_method(self, md_builder: System.Reflection.Emit.MethodBuilder) -> None:
        ...

    def set_add_on_method_core(self, md_builder: System.Reflection.Emit.MethodBuilder) -> None:
        """This method is protected."""
        ...

    @overload
    def set_custom_attribute(self, con: System.Reflection.ConstructorInfo, binary_attribute: typing.List[int]) -> None:
        ...

    @overload
    def set_custom_attribute(self, custom_builder: System.Reflection.Emit.CustomAttributeBuilder) -> None:
        ...

    def set_custom_attribute_core(self, con: System.Reflection.ConstructorInfo, binary_attribute: System.ReadOnlySpan[int]) -> None:
        """This method is protected."""
        ...

    def set_raise_method(self, md_builder: System.Reflection.Emit.MethodBuilder) -> None:
        ...

    def set_raise_method_core(self, md_builder: System.Reflection.Emit.MethodBuilder) -> None:
        """This method is protected."""
        ...

    def set_remove_on_method(self, md_builder: System.Reflection.Emit.MethodBuilder) -> None:
        ...

    def set_remove_on_method_core(self, md_builder: System.Reflection.Emit.MethodBuilder) -> None:
        """This method is protected."""
        ...


class PropertyBuilder(System.Reflection.PropertyInfo, metaclass=abc.ABCMeta):
    """This class has no documentation."""

    def __init__(self) -> None:
        """This method is protected."""
        ...

    def add_other_method(self, md_builder: System.Reflection.Emit.MethodBuilder) -> None:
        ...

    def add_other_method_core(self, md_builder: System.Reflection.Emit.MethodBuilder) -> None:
        """This method is protected."""
        ...

    def set_constant(self, default_value: typing.Any) -> None:
        ...

    def set_constant_core(self, default_value: typing.Any) -> None:
        """This method is protected."""
        ...

    @overload
    def set_custom_attribute(self, con: System.Reflection.ConstructorInfo, binary_attribute: typing.List[int]) -> None:
        ...

    @overload
    def set_custom_attribute(self, custom_builder: System.Reflection.Emit.CustomAttributeBuilder) -> None:
        ...

    def set_custom_attribute_core(self, con: System.Reflection.ConstructorInfo, binary_attribute: System.ReadOnlySpan[int]) -> None:
        """This method is protected."""
        ...

    def set_get_method(self, md_builder: System.Reflection.Emit.MethodBuilder) -> None:
        ...

    def set_get_method_core(self, md_builder: System.Reflection.Emit.MethodBuilder) -> None:
        """This method is protected."""
        ...

    def set_set_method(self, md_builder: System.Reflection.Emit.MethodBuilder) -> None:
        ...

    def set_set_method_core(self, md_builder: System.Reflection.Emit.MethodBuilder) -> None:
        """This method is protected."""
        ...


class TypeBuilder(System.Reflection.TypeInfo, metaclass=abc.ABCMeta):
    """This class has no documentation."""

    UNSPECIFIED_TYPE_SIZE: int = 0

    @property
    def packing_size(self) -> System.Reflection.Emit.PackingSize:
        ...

    @property
    @abc.abstractmethod
    def packing_size_core(self) -> System.Reflection.Emit.PackingSize:
        """This property is protected."""
        ...

    @property
    def size(self) -> int:
        ...

    @property
    @abc.abstractmethod
    def size_core(self) -> int:
        """This property is protected."""
        ...

    def __init__(self) -> None:
        """This method is protected."""
        ...

    def add_interface_implementation(self, interface_type: typing.Type) -> None:
        ...

    def add_interface_implementation_core(self, interface_type: typing.Type) -> None:
        """This method is protected."""
        ...

    def create_type(self) -> typing.Type:
        ...

    def create_type_info(self) -> System.Reflection.TypeInfo:
        ...

    def create_type_info_core(self) -> System.Reflection.TypeInfo:
        """This method is protected."""
        ...

    @overload
    def define_constructor(self, attributes: System.Reflection.MethodAttributes, calling_convention: System.Reflection.CallingConventions, parameter_types: typing.List[typing.Type]) -> System.Reflection.Emit.ConstructorBuilder:
        ...

    @overload
    def define_constructor(self, attributes: System.Reflection.MethodAttributes, calling_convention: System.Reflection.CallingConventions, parameter_types: typing.List[typing.Type], required_custom_modifiers: typing.List[typing.List[typing.Type]], optional_custom_modifiers: typing.List[typing.List[typing.Type]]) -> System.Reflection.Emit.ConstructorBuilder:
        ...

    def define_constructor_core(self, attributes: System.Reflection.MethodAttributes, calling_convention: System.Reflection.CallingConventions, parameter_types: typing.List[typing.Type], required_custom_modifiers: typing.List[typing.List[typing.Type]], optional_custom_modifiers: typing.List[typing.List[typing.Type]]) -> System.Reflection.Emit.ConstructorBuilder:
        """This method is protected."""
        ...

    def define_default_constructor(self, attributes: System.Reflection.MethodAttributes) -> System.Reflection.Emit.ConstructorBuilder:
        ...

    def define_default_constructor_core(self, attributes: System.Reflection.MethodAttributes) -> System.Reflection.Emit.ConstructorBuilder:
        """This method is protected."""
        ...

    def define_event(self, name: str, attributes: System.Reflection.EventAttributes, eventtype: typing.Type) -> System.Reflection.Emit.EventBuilder:
        ...

    def define_event_core(self, name: str, attributes: System.Reflection.EventAttributes, eventtype: typing.Type) -> System.Reflection.Emit.EventBuilder:
        """This method is protected."""
        ...

    @overload
    def define_field(self, field_name: str, type: typing.Type, attributes: System.Reflection.FieldAttributes) -> System.Reflection.Emit.FieldBuilder:
        ...

    @overload
    def define_field(self, field_name: str, type: typing.Type, required_custom_modifiers: typing.List[typing.Type], optional_custom_modifiers: typing.List[typing.Type], attributes: System.Reflection.FieldAttributes) -> System.Reflection.Emit.FieldBuilder:
        ...

    def define_field_core(self, field_name: str, type: typing.Type, required_custom_modifiers: typing.List[typing.Type], optional_custom_modifiers: typing.List[typing.Type], attributes: System.Reflection.FieldAttributes) -> System.Reflection.Emit.FieldBuilder:
        """This method is protected."""
        ...

    def define_generic_parameters(self, *names: str) -> typing.List[System.Reflection.Emit.GenericTypeParameterBuilder]:
        ...

    def define_generic_parameters_core(self, *names: str) -> typing.List[System.Reflection.Emit.GenericTypeParameterBuilder]:
        """This method is protected."""
        ...

    def define_initialized_data(self, name: str, data: typing.List[int], attributes: System.Reflection.FieldAttributes) -> System.Reflection.Emit.FieldBuilder:
        ...

    def define_initialized_data_core(self, name: str, data: typing.List[int], attributes: System.Reflection.FieldAttributes) -> System.Reflection.Emit.FieldBuilder:
        """This method is protected."""
        ...

    @overload
    def define_method(self, name: str, attributes: System.Reflection.MethodAttributes) -> System.Reflection.Emit.MethodBuilder:
        ...

    @overload
    def define_method(self, name: str, attributes: System.Reflection.MethodAttributes, calling_convention: System.Reflection.CallingConventions) -> System.Reflection.Emit.MethodBuilder:
        ...

    @overload
    def define_method(self, name: str, attributes: System.Reflection.MethodAttributes, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, parameter_types: typing.List[typing.Type]) -> System.Reflection.Emit.MethodBuilder:
        ...

    @overload
    def define_method(self, name: str, attributes: System.Reflection.MethodAttributes, return_type: typing.Type, parameter_types: typing.List[typing.Type]) -> System.Reflection.Emit.MethodBuilder:
        ...

    @overload
    def define_method(self, name: str, attributes: System.Reflection.MethodAttributes, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, return_type_required_custom_modifiers: typing.List[typing.Type], return_type_optional_custom_modifiers: typing.List[typing.Type], parameter_types: typing.List[typing.Type], parameter_type_required_custom_modifiers: typing.List[typing.List[typing.Type]], parameter_type_optional_custom_modifiers: typing.List[typing.List[typing.Type]]) -> System.Reflection.Emit.MethodBuilder:
        ...

    def define_method_core(self, name: str, attributes: System.Reflection.MethodAttributes, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, return_type_required_custom_modifiers: typing.List[typing.Type], return_type_optional_custom_modifiers: typing.List[typing.Type], parameter_types: typing.List[typing.Type], parameter_type_required_custom_modifiers: typing.List[typing.List[typing.Type]], parameter_type_optional_custom_modifiers: typing.List[typing.List[typing.Type]]) -> System.Reflection.Emit.MethodBuilder:
        """This method is protected."""
        ...

    def define_method_override(self, method_info_body: System.Reflection.MethodInfo, method_info_declaration: System.Reflection.MethodInfo) -> None:
        ...

    def define_method_override_core(self, method_info_body: System.Reflection.MethodInfo, method_info_declaration: System.Reflection.MethodInfo) -> None:
        """This method is protected."""
        ...

    @overload
    def define_nested_type(self, name: str) -> System.Reflection.Emit.TypeBuilder:
        ...

    @overload
    def define_nested_type(self, name: str, attr: System.Reflection.TypeAttributes) -> System.Reflection.Emit.TypeBuilder:
        ...

    @overload
    def define_nested_type(self, name: str, attr: System.Reflection.TypeAttributes, parent: typing.Type) -> System.Reflection.Emit.TypeBuilder:
        ...

    @overload
    def define_nested_type(self, name: str, attr: System.Reflection.TypeAttributes, parent: typing.Type, interfaces: typing.List[typing.Type]) -> System.Reflection.Emit.TypeBuilder:
        ...

    @overload
    def define_nested_type(self, name: str, attr: System.Reflection.TypeAttributes, parent: typing.Type, type_size: int) -> System.Reflection.Emit.TypeBuilder:
        ...

    @overload
    def define_nested_type(self, name: str, attr: System.Reflection.TypeAttributes, parent: typing.Type, pack_size: System.Reflection.Emit.PackingSize) -> System.Reflection.Emit.TypeBuilder:
        ...

    @overload
    def define_nested_type(self, name: str, attr: System.Reflection.TypeAttributes, parent: typing.Type, pack_size: System.Reflection.Emit.PackingSize, type_size: int) -> System.Reflection.Emit.TypeBuilder:
        ...

    def define_nested_type_core(self, name: str, attr: System.Reflection.TypeAttributes, parent: typing.Type, interfaces: typing.List[typing.Type], pack_size: System.Reflection.Emit.PackingSize, type_size: int) -> System.Reflection.Emit.TypeBuilder:
        """This method is protected."""
        ...

    @overload
    def define_p_invoke_method(self, name: str, dll_name: str, attributes: System.Reflection.MethodAttributes, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, parameter_types: typing.List[typing.Type], native_call_conv: System.Runtime.InteropServices.CallingConvention, native_char_set: System.Runtime.InteropServices.CharSet) -> System.Reflection.Emit.MethodBuilder:
        ...

    @overload
    def define_p_invoke_method(self, name: str, dll_name: str, entry_name: str, attributes: System.Reflection.MethodAttributes, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, parameter_types: typing.List[typing.Type], native_call_conv: System.Runtime.InteropServices.CallingConvention, native_char_set: System.Runtime.InteropServices.CharSet) -> System.Reflection.Emit.MethodBuilder:
        ...

    @overload
    def define_p_invoke_method(self, name: str, dll_name: str, entry_name: str, attributes: System.Reflection.MethodAttributes, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, return_type_required_custom_modifiers: typing.List[typing.Type], return_type_optional_custom_modifiers: typing.List[typing.Type], parameter_types: typing.List[typing.Type], parameter_type_required_custom_modifiers: typing.List[typing.List[typing.Type]], parameter_type_optional_custom_modifiers: typing.List[typing.List[typing.Type]], native_call_conv: System.Runtime.InteropServices.CallingConvention, native_char_set: System.Runtime.InteropServices.CharSet) -> System.Reflection.Emit.MethodBuilder:
        ...

    def define_p_invoke_method_core(self, name: str, dll_name: str, entry_name: str, attributes: System.Reflection.MethodAttributes, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, return_type_required_custom_modifiers: typing.List[typing.Type], return_type_optional_custom_modifiers: typing.List[typing.Type], parameter_types: typing.List[typing.Type], parameter_type_required_custom_modifiers: typing.List[typing.List[typing.Type]], parameter_type_optional_custom_modifiers: typing.List[typing.List[typing.Type]], native_call_conv: System.Runtime.InteropServices.CallingConvention, native_char_set: System.Runtime.InteropServices.CharSet) -> System.Reflection.Emit.MethodBuilder:
        """This method is protected."""
        ...

    @overload
    def define_property(self, name: str, attributes: System.Reflection.PropertyAttributes, return_type: typing.Type, parameter_types: typing.List[typing.Type]) -> System.Reflection.Emit.PropertyBuilder:
        ...

    @overload
    def define_property(self, name: str, attributes: System.Reflection.PropertyAttributes, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, parameter_types: typing.List[typing.Type]) -> System.Reflection.Emit.PropertyBuilder:
        ...

    @overload
    def define_property(self, name: str, attributes: System.Reflection.PropertyAttributes, return_type: typing.Type, return_type_required_custom_modifiers: typing.List[typing.Type], return_type_optional_custom_modifiers: typing.List[typing.Type], parameter_types: typing.List[typing.Type], parameter_type_required_custom_modifiers: typing.List[typing.List[typing.Type]], parameter_type_optional_custom_modifiers: typing.List[typing.List[typing.Type]]) -> System.Reflection.Emit.PropertyBuilder:
        ...

    @overload
    def define_property(self, name: str, attributes: System.Reflection.PropertyAttributes, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, return_type_required_custom_modifiers: typing.List[typing.Type], return_type_optional_custom_modifiers: typing.List[typing.Type], parameter_types: typing.List[typing.Type], parameter_type_required_custom_modifiers: typing.List[typing.List[typing.Type]], parameter_type_optional_custom_modifiers: typing.List[typing.List[typing.Type]]) -> System.Reflection.Emit.PropertyBuilder:
        ...

    def define_property_core(self, name: str, attributes: System.Reflection.PropertyAttributes, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, return_type_required_custom_modifiers: typing.List[typing.Type], return_type_optional_custom_modifiers: typing.List[typing.Type], parameter_types: typing.List[typing.Type], parameter_type_required_custom_modifiers: typing.List[typing.List[typing.Type]], parameter_type_optional_custom_modifiers: typing.List[typing.List[typing.Type]]) -> System.Reflection.Emit.PropertyBuilder:
        """This method is protected."""
        ...

    def define_type_initializer(self) -> System.Reflection.Emit.ConstructorBuilder:
        ...

    def define_type_initializer_core(self) -> System.Reflection.Emit.ConstructorBuilder:
        """This method is protected."""
        ...

    def define_uninitialized_data(self, name: str, size: int, attributes: System.Reflection.FieldAttributes) -> System.Reflection.Emit.FieldBuilder:
        ...

    def define_uninitialized_data_core(self, name: str, size: int, attributes: System.Reflection.FieldAttributes) -> System.Reflection.Emit.FieldBuilder:
        """This method is protected."""
        ...

    @staticmethod
    def get_constructor(type: typing.Type, constructor: System.Reflection.ConstructorInfo) -> System.Reflection.ConstructorInfo:
        ...

    @staticmethod
    def get_field(type: typing.Type, field: System.Reflection.FieldInfo) -> System.Reflection.FieldInfo:
        ...

    @staticmethod
    def get_method(type: typing.Type, method: System.Reflection.MethodInfo) -> System.Reflection.MethodInfo:
        ...

    def is_created(self) -> bool:
        ...

    def is_created_core(self) -> bool:
        """This method is protected."""
        ...

    @overload
    def make_array_type(self) -> typing.Type:
        ...

    @overload
    def make_array_type(self, rank: int) -> typing.Type:
        ...

    def make_by_ref_type(self) -> typing.Type:
        ...

    def make_generic_type(self, *type_arguments: typing.Type) -> typing.Type:
        ...

    def make_pointer_type(self) -> typing.Type:
        ...

    @overload
    def set_custom_attribute(self, con: System.Reflection.ConstructorInfo, binary_attribute: typing.List[int]) -> None:
        ...

    @overload
    def set_custom_attribute(self, custom_builder: System.Reflection.Emit.CustomAttributeBuilder) -> None:
        ...

    def set_custom_attribute_core(self, con: System.Reflection.ConstructorInfo, binary_attribute: System.ReadOnlySpan[int]) -> None:
        """This method is protected."""
        ...

    def set_parent(self, parent: typing.Type) -> None:
        ...

    def set_parent_core(self, parent: typing.Type) -> None:
        """This method is protected."""
        ...


class ModuleBuilder(System.Reflection.Module, metaclass=abc.ABCMeta):
    """This class has no documentation."""

    def __init__(self) -> None:
        """This method is protected."""
        ...

    def create_global_functions(self) -> None:
        ...

    def create_global_functions_core(self) -> None:
        """This method is protected."""
        ...

    @overload
    def define_document(self, url: str, language: System.Guid, language_vendor: System.Guid, document_type: System.Guid) -> System.Diagnostics.SymbolStore.ISymbolDocumentWriter:
        """
        Defines a document for source.
        
        :param url: The URL for the document.
        :param language: The GUID that identifies the document language. This can be Empty
        :param language_vendor: The GUID that identifies the document language vendor. This is not used.
        :param document_type: The GUID that identifies the document language vendor. This is not used.
        :returns: The defined document.
        """
        ...

    @overload
    def define_document(self, url: str, language: System.Guid = ...) -> System.Diagnostics.SymbolStore.ISymbolDocumentWriter:
        """
        Defines a document for source.
        
        :param url: The URL for the document.
        :param language: The GUID that identifies the document language. This is optional
        :returns: The defined document.
        """
        ...

    def define_document_core(self, url: str, language: System.Guid = ...) -> System.Diagnostics.SymbolStore.ISymbolDocumentWriter:
        """
        When override in a derived class, defines a document for source.
        
        This method is protected.
        
        :param url: The URL for the document.
        :param language: The GUID that identifies the document language. This is optional
        :returns: The defined document.
        """
        ...

    def define_enum(self, name: str, visibility: System.Reflection.TypeAttributes, underlying_type: typing.Type) -> System.Reflection.Emit.EnumBuilder:
        ...

    def define_enum_core(self, name: str, visibility: System.Reflection.TypeAttributes, underlying_type: typing.Type) -> System.Reflection.Emit.EnumBuilder:
        """This method is protected."""
        ...

    @overload
    def define_global_method(self, name: str, attributes: System.Reflection.MethodAttributes, return_type: typing.Type, parameter_types: typing.List[typing.Type]) -> System.Reflection.Emit.MethodBuilder:
        ...

    @overload
    def define_global_method(self, name: str, attributes: System.Reflection.MethodAttributes, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, parameter_types: typing.List[typing.Type]) -> System.Reflection.Emit.MethodBuilder:
        ...

    @overload
    def define_global_method(self, name: str, attributes: System.Reflection.MethodAttributes, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, required_return_type_custom_modifiers: typing.List[typing.Type], optional_return_type_custom_modifiers: typing.List[typing.Type], parameter_types: typing.List[typing.Type], required_parameter_type_custom_modifiers: typing.List[typing.List[typing.Type]], optional_parameter_type_custom_modifiers: typing.List[typing.List[typing.Type]]) -> System.Reflection.Emit.MethodBuilder:
        ...

    def define_global_method_core(self, name: str, attributes: System.Reflection.MethodAttributes, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, required_return_type_custom_modifiers: typing.List[typing.Type], optional_return_type_custom_modifiers: typing.List[typing.Type], parameter_types: typing.List[typing.Type], required_parameter_type_custom_modifiers: typing.List[typing.List[typing.Type]], optional_parameter_type_custom_modifiers: typing.List[typing.List[typing.Type]]) -> System.Reflection.Emit.MethodBuilder:
        """This method is protected."""
        ...

    def define_initialized_data(self, name: str, data: typing.List[int], attributes: System.Reflection.FieldAttributes) -> System.Reflection.Emit.FieldBuilder:
        ...

    def define_initialized_data_core(self, name: str, data: typing.List[int], attributes: System.Reflection.FieldAttributes) -> System.Reflection.Emit.FieldBuilder:
        """This method is protected."""
        ...

    @overload
    def define_p_invoke_method(self, name: str, dll_name: str, attributes: System.Reflection.MethodAttributes, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, parameter_types: typing.List[typing.Type], native_call_conv: System.Runtime.InteropServices.CallingConvention, native_char_set: System.Runtime.InteropServices.CharSet) -> System.Reflection.Emit.MethodBuilder:
        ...

    @overload
    def define_p_invoke_method(self, name: str, dll_name: str, entry_name: str, attributes: System.Reflection.MethodAttributes, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, parameter_types: typing.List[typing.Type], native_call_conv: System.Runtime.InteropServices.CallingConvention, native_char_set: System.Runtime.InteropServices.CharSet) -> System.Reflection.Emit.MethodBuilder:
        ...

    def define_p_invoke_method_core(self, name: str, dll_name: str, entry_name: str, attributes: System.Reflection.MethodAttributes, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, parameter_types: typing.List[typing.Type], native_call_conv: System.Runtime.InteropServices.CallingConvention, native_char_set: System.Runtime.InteropServices.CharSet) -> System.Reflection.Emit.MethodBuilder:
        """This method is protected."""
        ...

    @overload
    def define_type(self, name: str) -> System.Reflection.Emit.TypeBuilder:
        ...

    @overload
    def define_type(self, name: str, attr: System.Reflection.TypeAttributes) -> System.Reflection.Emit.TypeBuilder:
        ...

    @overload
    def define_type(self, name: str, attr: System.Reflection.TypeAttributes, parent: typing.Type) -> System.Reflection.Emit.TypeBuilder:
        ...

    @overload
    def define_type(self, name: str, attr: System.Reflection.TypeAttributes, parent: typing.Type, interfaces: typing.List[typing.Type]) -> System.Reflection.Emit.TypeBuilder:
        ...

    @overload
    def define_type(self, name: str, attr: System.Reflection.TypeAttributes, parent: typing.Type, typesize: int) -> System.Reflection.Emit.TypeBuilder:
        ...

    @overload
    def define_type(self, name: str, attr: System.Reflection.TypeAttributes, parent: typing.Type, packsize: System.Reflection.Emit.PackingSize) -> System.Reflection.Emit.TypeBuilder:
        ...

    @overload
    def define_type(self, name: str, attr: System.Reflection.TypeAttributes, parent: typing.Type, packing_size: System.Reflection.Emit.PackingSize, typesize: int) -> System.Reflection.Emit.TypeBuilder:
        ...

    def define_type_core(self, name: str, attr: System.Reflection.TypeAttributes, parent: typing.Type, interfaces: typing.List[typing.Type], packing_size: System.Reflection.Emit.PackingSize, typesize: int) -> System.Reflection.Emit.TypeBuilder:
        """This method is protected."""
        ...

    def define_uninitialized_data(self, name: str, size: int, attributes: System.Reflection.FieldAttributes) -> System.Reflection.Emit.FieldBuilder:
        ...

    def define_uninitialized_data_core(self, name: str, size: int, attributes: System.Reflection.FieldAttributes) -> System.Reflection.Emit.FieldBuilder:
        """This method is protected."""
        ...

    def get_array_method(self, array_class: typing.Type, method_name: str, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, parameter_types: typing.List[typing.Type]) -> System.Reflection.MethodInfo:
        ...

    def get_array_method_core(self, array_class: typing.Type, method_name: str, calling_convention: System.Reflection.CallingConventions, return_type: typing.Type, parameter_types: typing.List[typing.Type]) -> System.Reflection.MethodInfo:
        """This method is protected."""
        ...

    def get_field_metadata_token(self, field: System.Reflection.FieldInfo) -> int:
        ...

    @overload
    def get_method_metadata_token(self, method: System.Reflection.MethodInfo) -> int:
        ...

    @overload
    def get_method_metadata_token(self, constructor: System.Reflection.ConstructorInfo) -> int:
        ...

    def get_signature_metadata_token(self, signature: System.Reflection.Emit.SignatureHelper) -> int:
        ...

    def get_string_metadata_token(self, string_constant: str) -> int:
        ...

    def get_type_metadata_token(self, type: typing.Type) -> int:
        ...

    @overload
    def set_custom_attribute(self, con: System.Reflection.ConstructorInfo, binary_attribute: typing.List[int]) -> None:
        ...

    @overload
    def set_custom_attribute(self, custom_builder: System.Reflection.Emit.CustomAttributeBuilder) -> None:
        ...

    def set_custom_attribute_core(self, con: System.Reflection.ConstructorInfo, binary_attribute: System.ReadOnlySpan[int]) -> None:
        """This method is protected."""
        ...


class AssemblyBuilderAccess(Enum):
    """This class has no documentation."""

    RUN = 1

    RUN_AND_COLLECT = ...


class AssemblyBuilder(System.Reflection.Assembly, metaclass=abc.ABCMeta):
    """This class has no documentation."""

    @property
    def code_base(self) -> str:
        """Assembly.CodeBase and Assembly.EscapedCodeBase are only included for .NET Framework compatibility. Use Assembly.Location instead."""
        warnings.warn("Assembly.CodeBase and Assembly.EscapedCodeBase are only included for .NET Framework compatibility. Use Assembly.Location instead.", DeprecationWarning)

    @property
    def location(self) -> str:
        ...

    @property
    def entry_point(self) -> System.Reflection.MethodInfo:
        ...

    @property
    def is_dynamic(self) -> bool:
        ...

    def __init__(self) -> None:
        """This method is protected."""
        ...

    @staticmethod
    @overload
    def define_dynamic_assembly(name: System.Reflection.AssemblyName, access: System.Reflection.Emit.AssemblyBuilderAccess) -> System.Reflection.Emit.AssemblyBuilder:
        ...

    @staticmethod
    @overload
    def define_dynamic_assembly(name: System.Reflection.AssemblyName, access: System.Reflection.Emit.AssemblyBuilderAccess, assembly_attributes: System.Collections.Generic.IEnumerable[System.Reflection.Emit.CustomAttributeBuilder]) -> System.Reflection.Emit.AssemblyBuilder:
        ...

    def define_dynamic_module(self, name: str) -> System.Reflection.Emit.ModuleBuilder:
        ...

    def define_dynamic_module_core(self, name: str) -> System.Reflection.Emit.ModuleBuilder:
        """This method is protected."""
        ...

    def get_dynamic_module(self, name: str) -> System.Reflection.Emit.ModuleBuilder:
        ...

    def get_dynamic_module_core(self, name: str) -> System.Reflection.Emit.ModuleBuilder:
        """This method is protected."""
        ...

    def get_exported_types(self) -> typing.List[typing.Type]:
        ...

    def get_file(self, name: str) -> System.IO.FileStream:
        ...

    def get_files(self, get_resource_modules: bool) -> typing.List[System.IO.FileStream]:
        ...

    def get_manifest_resource_info(self, resource_name: str) -> System.Reflection.ManifestResourceInfo:
        ...

    def get_manifest_resource_names(self) -> typing.List[str]:
        ...

    @overload
    def get_manifest_resource_stream(self, name: str) -> System.IO.Stream:
        ...

    @overload
    def get_manifest_resource_stream(self, type: typing.Type, name: str) -> System.IO.Stream:
        ...

    @overload
    def set_custom_attribute(self, con: System.Reflection.ConstructorInfo, binary_attribute: typing.List[int]) -> None:
        ...

    @overload
    def set_custom_attribute(self, custom_builder: System.Reflection.Emit.CustomAttributeBuilder) -> None:
        ...

    def set_custom_attribute_core(self, con: System.Reflection.ConstructorInfo, binary_attribute: System.ReadOnlySpan[int]) -> None:
        """This method is protected."""
        ...


class PEFileKinds(Enum):
    """This class has no documentation."""

    DLL = ...

    CONSOLE_APPLICATION = ...

    WINDOW_APPLICATION = ...


class DynamicMethod(System.Reflection.MethodInfo):
    """This class has no documentation."""

    @property
    def name(self) -> str:
        ...

    @property
    def declaring_type(self) -> typing.Type:
        ...

    @property
    def reflected_type(self) -> typing.Type:
        ...

    @property
    def module(self) -> System.Reflection.Module:
        ...

    @property
    def method_handle(self) -> System.RuntimeMethodHandle:
        ...

    @property
    def attributes(self) -> System.Reflection.MethodAttributes:
        ...

    @property
    def calling_convention(self) -> System.Reflection.CallingConventions:
        ...

    @property
    def is_security_critical(self) -> bool:
        ...

    @property
    def is_security_safe_critical(self) -> bool:
        ...

    @property
    def is_security_transparent(self) -> bool:
        ...

    @property
    def return_type(self) -> typing.Type:
        ...

    @property
    def return_parameter(self) -> System.Reflection.ParameterInfo:
        ...

    @property
    def return_type_custom_attributes(self) -> System.Reflection.ICustomAttributeProvider:
        ...

    @property
    def init_locals(self) -> bool:
        ...

    @property.setter
    def init_locals(self, value: bool) -> None:
        ...

    @overload
    def __init__(self, name: str, returnType: typing.Type, parameterTypes: typing.List[typing.Type]) -> None:
        ...

    @overload
    def __init__(self, name: str, returnType: typing.Type, parameterTypes: typing.List[typing.Type], restrictedSkipVisibility: bool) -> None:
        ...

    @overload
    def __init__(self, name: str, returnType: typing.Type, parameterTypes: typing.List[typing.Type], m: System.Reflection.Module) -> None:
        ...

    @overload
    def __init__(self, name: str, returnType: typing.Type, parameterTypes: typing.List[typing.Type], m: System.Reflection.Module, skipVisibility: bool) -> None:
        ...

    @overload
    def __init__(self, name: str, attributes: System.Reflection.MethodAttributes, callingConvention: System.Reflection.CallingConventions, returnType: typing.Type, parameterTypes: typing.List[typing.Type], m: System.Reflection.Module, skipVisibility: bool) -> None:
        ...

    @overload
    def __init__(self, name: str, returnType: typing.Type, parameterTypes: typing.List[typing.Type], owner: typing.Type) -> None:
        ...

    @overload
    def __init__(self, name: str, returnType: typing.Type, parameterTypes: typing.List[typing.Type], owner: typing.Type, skipVisibility: bool) -> None:
        ...

    @overload
    def __init__(self, name: str, attributes: System.Reflection.MethodAttributes, callingConvention: System.Reflection.CallingConventions, returnType: typing.Type, parameterTypes: typing.List[typing.Type], owner: typing.Type, skipVisibility: bool) -> None:
        ...

    @overload
    def create_delegate(self, delegate_type: typing.Type) -> System.Delegate:
        ...

    @overload
    def create_delegate(self, delegate_type: typing.Type, target: typing.Any) -> System.Delegate:
        ...

    def define_parameter(self, position: int, attributes: System.Reflection.ParameterAttributes, parameter_name: str) -> System.Reflection.Emit.ParameterBuilder:
        ...

    def get_base_definition(self) -> System.Reflection.MethodInfo:
        ...

    @overload
    def get_custom_attributes(self, attribute_type: typing.Type, inherit: bool) -> typing.List[System.Object]:
        ...

    @overload
    def get_custom_attributes(self, inherit: bool) -> typing.List[System.Object]:
        ...

    def get_dynamic_il_info(self) -> System.Reflection.Emit.DynamicILInfo:
        ...

    @overload
    def get_il_generator(self) -> System.Reflection.Emit.ILGenerator:
        ...

    @overload
    def get_il_generator(self, stream_size: int) -> System.Reflection.Emit.ILGenerator:
        ...

    def get_method_implementation_flags(self) -> System.Reflection.MethodImplAttributes:
        ...

    def get_parameters(self) -> typing.List[System.Reflection.ParameterInfo]:
        ...

    def invoke(self, obj: typing.Any, invoke_attr: System.Reflection.BindingFlags, binder: System.Reflection.Binder, parameters: typing.List[System.Object], culture: System.Globalization.CultureInfo) -> System.Object:
        ...

    def is_defined(self, attribute_type: typing.Type, inherit: bool) -> bool:
        ...

    def to_string(self) -> str:
        ...





from typing import overload
from enum import Enum
import abc
import datetime
import typing
import warnings

import System
import System.Collections
import System.Runtime.Serialization


class DeserializationToken(System.IDisposable):
    """This class has no documentation."""

    def dispose(self) -> None:
        ...


class IFormatterConverter(metaclass=abc.ABCMeta):
    """Obsoletions.LegacyFormatterMessage"""

    @overload
    def convert(self, value: typing.Any, type: typing.Type) -> System.Object:
        ...

    @overload
    def convert(self, value: typing.Any, type_code: System.TypeCode) -> System.Object:
        ...

    def to_boolean(self, value: typing.Any) -> bool:
        ...

    def to_byte(self, value: typing.Any) -> int:
        ...

    def to_char(self, value: typing.Any) -> str:
        ...

    def to_date_time(self, value: typing.Any) -> datetime.datetime:
        ...

    def to_decimal(self, value: typing.Any) -> float:
        ...

    def to_double(self, value: typing.Any) -> float:
        ...

    def to_int_16(self, value: typing.Any) -> int:
        ...

    def to_int_32(self, value: typing.Any) -> int:
        ...

    def to_int_64(self, value: typing.Any) -> int:
        ...

    def to_s_byte(self, value: typing.Any) -> int:
        ...

    def to_single(self, value: typing.Any) -> float:
        ...

    def to_string(self, value: typing.Any) -> str:
        ...

    def to_u_int_16(self, value: typing.Any) -> int:
        ...

    def to_u_int_32(self, value: typing.Any) -> int:
        ...

    def to_u_int_64(self, value: typing.Any) -> int:
        ...


class SerializationEntry:
    """This class has no documentation."""

    @property
    def value(self) -> System.Object:
        ...

    @property
    def name(self) -> str:
        ...

    @property
    def object_type(self) -> typing.Type:
        ...


class SerializationInfoEnumerator(System.Object, System.Collections.IEnumerator):
    """This class has no documentation."""

    @property
    def current(self) -> System.Runtime.Serialization.SerializationEntry:
        ...

    @property
    def name(self) -> str:
        ...

    @property
    def value(self) -> System.Object:
        ...

    @property
    def object_type(self) -> typing.Type:
        ...

    def move_next(self) -> bool:
        ...

    def reset(self) -> None:
        ...


class SerializationInfo(System.Object):
    """The structure for holding all of the data needed for object serialization and deserialization."""

    @property
    def full_type_name(self) -> str:
        ...

    @property.setter
    def full_type_name(self, value: str) -> None:
        ...

    @property
    def assembly_name(self) -> str:
        ...

    @property.setter
    def assembly_name(self, value: str) -> None:
        ...

    @property
    def is_full_type_name_set_explicit(self) -> bool:
        ...

    @property
    def is_assembly_name_set_explicit(self) -> bool:
        ...

    @property
    def member_count(self) -> int:
        ...

    @property
    def object_type(self) -> typing.Type:
        ...

    @overload
    def __init__(self, type: typing.Type, converter: System.Runtime.Serialization.IFormatterConverter) -> None:
        """Obsoletions.LegacyFormatterMessage"""
        ...

    @overload
    def __init__(self, type: typing.Type, converter: System.Runtime.Serialization.IFormatterConverter, requireSameTokenInPartialTrust: bool) -> None:
        """Obsoletions.LegacyFormatterMessage"""
        ...

    @overload
    def add_value(self, name: str, value: typing.Any, type: typing.Type) -> None:
        ...

    @overload
    def add_value(self, name: str, value: typing.Any) -> None:
        ...

    @overload
    def add_value(self, name: str, value: bool) -> None:
        ...

    @overload
    def add_value(self, name: str, value: str) -> None:
        ...

    @overload
    def add_value(self, name: str, value: int) -> None:
        ...

    @overload
    def add_value(self, name: str, value: int) -> None:
        ...

    @overload
    def add_value(self, name: str, value: int) -> None:
        ...

    @overload
    def add_value(self, name: str, value: int) -> None:
        ...

    @overload
    def add_value(self, name: str, value: int) -> None:
        ...

    @overload
    def add_value(self, name: str, value: int) -> None:
        ...

    @overload
    def add_value(self, name: str, value: int) -> None:
        ...

    @overload
    def add_value(self, name: str, value: int) -> None:
        ...

    @overload
    def add_value(self, name: str, value: float) -> None:
        ...

    @overload
    def add_value(self, name: str, value: float) -> None:
        ...

    @overload
    def add_value(self, name: str, value: float) -> None:
        ...

    @overload
    def add_value(self, name: str, value: typing.Union[datetime.datetime, datetime.date]) -> None:
        ...

    def get_boolean(self, name: str) -> bool:
        ...

    def get_byte(self, name: str) -> int:
        ...

    def get_char(self, name: str) -> str:
        ...

    def get_date_time(self, name: str) -> datetime.datetime:
        ...

    def get_decimal(self, name: str) -> float:
        ...

    def get_double(self, name: str) -> float:
        ...

    def get_enumerator(self) -> System.Runtime.Serialization.SerializationInfoEnumerator:
        ...

    def get_int_16(self, name: str) -> int:
        ...

    def get_int_32(self, name: str) -> int:
        ...

    def get_int_64(self, name: str) -> int:
        ...

    def get_s_byte(self, name: str) -> int:
        ...

    def get_single(self, name: str) -> float:
        ...

    def get_string(self, name: str) -> str:
        ...

    def get_u_int_16(self, name: str) -> int:
        ...

    def get_u_int_32(self, name: str) -> int:
        ...

    def get_u_int_64(self, name: str) -> int:
        ...

    def get_value(self, name: str, type: typing.Type) -> System.Object:
        ...

    def set_type(self, type: typing.Type) -> None:
        ...

    @staticmethod
    def start_deserialization() -> System.Runtime.Serialization.DeserializationToken:
        ...


class StreamingContextStates(Enum):
    """Obsoletions.LegacyFormatterMessage"""

    CROSS_PROCESS = ...

    CROSS_MACHINE = ...

    FILE = ...

    PERSISTENCE = ...

    REMOTING = ...

    OTHER = ...

    CLONE = ...

    CROSS_APP_DOMAIN = ...

    ALL = ...


class StreamingContext:
    """This class has no documentation."""

    @property
    def state(self) -> System.Runtime.Serialization.StreamingContextStates:
        """Obsoletions.LegacyFormatterMessage"""
        warnings.warn("Obsoletions.LegacyFormatterMessage", DeprecationWarning)

    @property
    def context(self) -> System.Object:
        ...

    @overload
    def __init__(self, state: System.Runtime.Serialization.StreamingContextStates) -> None:
        """Obsoletions.LegacyFormatterMessage"""
        ...

    @overload
    def __init__(self, state: System.Runtime.Serialization.StreamingContextStates, additional: typing.Any) -> None:
        """Obsoletions.LegacyFormatterMessage"""
        ...

    def equals(self, obj: typing.Any) -> bool:
        ...

    def get_hash_code(self) -> int:
        ...


class SerializationException(System.SystemException):
    """This class has no documentation."""

    @overload
    def __init__(self) -> None:
        """
        Creates a new SerializationException with its message
        string set to a default message.
        """
        ...

    @overload
    def __init__(self, message: str) -> None:
        ...

    @overload
    def __init__(self, message: str, innerException: System.Exception) -> None:
        ...

    @overload
    def __init__(self, info: System.Runtime.Serialization.SerializationInfo, context: System.Runtime.Serialization.StreamingContext) -> None:
        """
        This method is protected.
        
        Obsoletions.LegacyFormatterImplMessage
        """
        ...


class ISafeSerializationData(metaclass=abc.ABCMeta):
    """Obsoletions.LegacyFormatterMessage"""

    def complete_deserialization(self, deserialized: typing.Any) -> None:
        ...


class SafeSerializationEventArgs(System.EventArgs):
    """Obsoletions.LegacyFormatterMessage"""

    @property
    def streaming_context(self) -> System.Runtime.Serialization.StreamingContext:
        ...

    def add_serialized_state(self, serialized_state: System.Runtime.Serialization.ISafeSerializationData) -> None:
        ...


class OnDeserializedAttribute(System.Attribute):
    """This class has no documentation."""


class IDeserializationCallback(metaclass=abc.ABCMeta):
    """This class has no documentation."""

    def on_deserialization(self, sender: typing.Any) -> None:
        ...


class ISerializable(metaclass=abc.ABCMeta):
    """This class has no documentation."""

    def get_object_data(self, info: System.Runtime.Serialization.SerializationInfo, context: System.Runtime.Serialization.StreamingContext) -> None:
        """Obsoletions.LegacyFormatterMessage"""
        warnings.warn("Obsoletions.LegacyFormatterMessage", DeprecationWarning)


class OnSerializedAttribute(System.Attribute):
    """This class has no documentation."""


class OnDeserializingAttribute(System.Attribute):
    """This class has no documentation."""


class IObjectReference(metaclass=abc.ABCMeta):
    """Obsoletions.LegacyFormatterMessage"""

    def get_real_object(self, context: System.Runtime.Serialization.StreamingContext) -> System.Object:
        ...


class OptionalFieldAttribute(System.Attribute):
    """This class has no documentation."""

    @property
    def version_added(self) -> int:
        ...

    @property.setter
    def version_added(self, value: int) -> None:
        ...


class OnSerializingAttribute(System.Attribute):
    """This class has no documentation."""





from typing import overload
from enum import Enum
import abc
import datetime
import typing
import warnings

import QuantConnect
import QuantConnect.Data.UniverseSelection
import QuantConnect.Interfaces
import QuantConnect.Lean.Engine.RealTime
import QuantConnect.Lean.Engine.Results
import QuantConnect.Packets
import QuantConnect.Scheduling
import QuantConnect.Securities
import System
import System.Collections.Concurrent
import System.Collections.Generic


class IRealTimeHandler(QuantConnect.Scheduling.IEventSchedule, metaclass=abc.ABCMeta):
    """Real time event handler, trigger functions at regular or pretimed intervals"""

    @property
    @abc.abstractmethod
    def is_active(self) -> bool:
        """Thread status flag."""
        ...

    def exit(self) -> None:
        """Trigger and exit signal to terminate real time event scanner."""
        ...

    def on_securities_changed(self, changes: QuantConnect.Data.UniverseSelection.SecurityChanges) -> None:
        """Event fired each time that we add/remove securities from the data feed"""
        ...

    def scan_past_events(self, time: typing.Union[datetime.datetime, datetime.date]) -> None:
        """
        Scan for past events that didn't fire because there was no data at the scheduled time.
        
        :param time: Current time.
        """
        ...

    def set_time(self, time: typing.Union[datetime.datetime, datetime.date]) -> None:
        """
        Set the current time for the event scanner (so we can use same code for backtesting and live events)
        
        :param time: Current real or backtest time.
        """
        ...

    def setup(self, algorithm: QuantConnect.Interfaces.IAlgorithm, job: QuantConnect.Packets.AlgorithmNodePacket, result_handler: QuantConnect.Lean.Engine.Results.IResultHandler, api: QuantConnect.Interfaces.IApi, isolator_limit_provider: QuantConnect.IIsolatorLimitResultProvider) -> None:
        """Initializes the real time handler for the specified algorithm and job"""
        ...


class BaseRealTimeHandler(System.Object, QuantConnect.Lean.Engine.RealTime.IRealTimeHandler, metaclass=abc.ABCMeta):
    """
    Base class for the real time handler LiveTradingRealTimeHandler
    and BacktestingRealTimeHandler implementations
    """

    @property
    @abc.abstractmethod
    def is_active(self) -> bool:
        """Thread status flag."""
        ...

    @property
    def scheduled_events(self) -> System.Collections.Concurrent.ConcurrentDictionary[QuantConnect.Scheduling.ScheduledEvent, int]:
        """
        The scheduled events container
        
        This property is protected.
        """
        ...

    @property
    def isolator_limit_provider(self) -> QuantConnect.IIsolatorLimitResultProvider:
        """
        The isolator limit result provider instance
        
        This property is protected.
        """
        ...

    @property
    def algorithm(self) -> QuantConnect.Interfaces.IAlgorithm:
        """
        The algorithm instance
        
        This property is protected.
        """
        ...

    @property
    def time_monitor(self) -> QuantConnect.Scheduling.TimeMonitor:
        """
        The time monitor instance to use
        
        This property is protected.
        """
        ...

    def add(self, scheduled_event: QuantConnect.Scheduling.ScheduledEvent) -> None:
        """
        Adds the specified event to the schedule
        
        :param scheduled_event: The event to be scheduled, including the date/times the event fires and the callback
        """
        ...

    def exit(self) -> None:
        """Stop the real time thread"""
        ...

    def get_scheduled_event_unique_id(self) -> int:
        """
        Gets a new scheduled event unique id
        
        This method is protected.
        """
        ...

    def get_time_monitor_timeout(self) -> int:
        """
        Get's the timeout the scheduled task time monitor should use
        
        This method is protected.
        """
        ...

    def on_securities_changed(self, changes: QuantConnect.Data.UniverseSelection.SecurityChanges) -> None:
        """Event fired each time that we add/remove securities from the data feed"""
        ...

    def remove(self, scheduled_event: QuantConnect.Scheduling.ScheduledEvent) -> None:
        """
        Removes the specified event from the schedule
        
        :param scheduled_event: The event to be removed
        """
        ...

    def scan_past_events(self, time: typing.Union[datetime.datetime, datetime.date]) -> None:
        """
        Scan for past events that didn't fire because there was no data at the scheduled time.
        
        :param time: Current time.
        """
        ...

    def set_time(self, time: typing.Union[datetime.datetime, datetime.date]) -> None:
        """
        Set the current time for the event scanner (so we can use same code for backtesting and live events)
        
        :param time: Current real or backtest time.
        """
        ...

    def setup(self, algorithm: QuantConnect.Interfaces.IAlgorithm, job: QuantConnect.Packets.AlgorithmNodePacket, result_handler: QuantConnect.Lean.Engine.Results.IResultHandler, api: QuantConnect.Interfaces.IApi, isolator_limit_provider: QuantConnect.IIsolatorLimitResultProvider) -> None:
        """
        Initializes the real time handler for the specified algorithm and job.
        Adds EndOfDayEvents
        """
        ...


class BacktestingRealTimeHandler(QuantConnect.Lean.Engine.RealTime.BaseRealTimeHandler):
    """Pseudo realtime event processing for backtesting to simulate realtime events in fast forward."""

    @property
    def is_active(self) -> bool:
        """
        Flag indicating the hander thread is completely finished and ready to dispose.
        this doesn't run as its own thread
        """
        ...

    def add(self, scheduled_event: QuantConnect.Scheduling.ScheduledEvent) -> None:
        """
        Adds the specified event to the schedule
        
        :param scheduled_event: The event to be scheduled, including the date/times the event fires and the callback
        """
        ...

    def remove(self, scheduled_event: QuantConnect.Scheduling.ScheduledEvent) -> None:
        """
        Removes the specified event from the schedule
        
        :param scheduled_event: The event to be removed
        """
        ...

    def scan_past_events(self, time: typing.Union[datetime.datetime, datetime.date]) -> None:
        """
        Scan for past events that didn't fire because there was no data at the scheduled time.
        
        :param time: Current time.
        """
        ...

    def set_time(self, time: typing.Union[datetime.datetime, datetime.date]) -> None:
        """
        Set the time for the realtime event handler.
        
        :param time: Current time.
        """
        ...

    def setup(self, algorithm: QuantConnect.Interfaces.IAlgorithm, job: QuantConnect.Packets.AlgorithmNodePacket, result_handler: QuantConnect.Lean.Engine.Results.IResultHandler, api: QuantConnect.Interfaces.IApi, isolator_limit_provider: QuantConnect.IIsolatorLimitResultProvider) -> None:
        """Initializes the real time handler for the specified algorithm and job"""
        ...

    @staticmethod
    def sort_first_element(scheduled_events: System.Collections.Generic.IList[QuantConnect.Scheduling.ScheduledEvent]) -> None:
        """
        Sorts the first element of the provided list and supposes the rest of the collection is sorted.
        Supposes the collection has at least 1 element
        """
        ...


class LiveTradingRealTimeHandler(QuantConnect.Lean.Engine.RealTime.BacktestingRealTimeHandler):
    """Live trading realtime event processing."""

    @property
    def market_hours_database(self) -> QuantConnect.Securities.MarketHoursDatabase:
        """
        Gets the current market hours database instance
        
        This property is protected.
        """
        ...

    @property.setter
    def market_hours_database(self, value: QuantConnect.Securities.MarketHoursDatabase) -> None:
        ...

    @property
    def symbol_properties_database(self) -> QuantConnect.Securities.SymbolPropertiesDatabase:
        """
        Gets the current symbol properties database instance
        
        This property is protected.
        """
        ...

    @property.setter
    def symbol_properties_database(self, value: QuantConnect.Securities.SymbolPropertiesDatabase) -> None:
        ...

    @property
    def time_provider(self) -> QuantConnect.ITimeProvider:
        """
        Gets the time provider
        
        This property is protected.
        """
        ...

    @property
    def is_active(self) -> bool:
        """Boolean flag indicating thread state."""
        ...

    def exit(self) -> None:
        """Stop the real time thread"""
        ...

    def get_time_monitor_timeout(self) -> int:
        """
        Get's the timeout the scheduled task time monitor should use
        
        This method is protected.
        """
        ...

    def reset_market_hours_database(self) -> None:
        """
        Resets the market hours database, forcing a reload when reused.
        Called in tests where multiple algorithms are run sequentially,
        and we need to guarantee that every test starts with the same environment.
        
        This method is protected.
        """
        ...

    def reset_symbol_properties_database(self) -> None:
        """
        Resets the symbol properties database, forcing a reload when reused.
        
        This method is protected.
        """
        ...

    def scan_past_events(self, time: typing.Union[datetime.datetime, datetime.date]) -> None:
        """
        Scan for past events that didn't fire because there was no data at the scheduled time.
        
        :param time: Current time.
        """
        ...

    def set_time(self, time: typing.Union[datetime.datetime, datetime.date]) -> None:
        """Set the current time. If the date changes re-start the realtime event setup routines."""
        ...

    def setup(self, algorithm: QuantConnect.Interfaces.IAlgorithm, job: QuantConnect.Packets.AlgorithmNodePacket, result_handler: QuantConnect.Lean.Engine.Results.IResultHandler, api: QuantConnect.Interfaces.IApi, isolator_limit_provider: QuantConnect.IIsolatorLimitResultProvider) -> None:
        """Initializes the real time handler for the specified algorithm and job"""
        ...


class ScheduledEventFactory(System.Object):
    """Provides methods for creating common scheduled events"""

    @staticmethod
    def create_event_name(scope: str, name: str) -> str:
        """
        Defines the format of event names generated by this system.
        
        :param scope: The scope of the event, example, 'Algorithm' or 'Security'
        :param name: A name for this specified event in this scope, example, 'EndOfDay'
        :returns: A string representing a fully scoped event name.
        """
        ...

    @staticmethod
    def every_algorithm_end_of_day(algorithm: QuantConnect.Interfaces.IAlgorithm, result_handler: QuantConnect.Lean.Engine.Results.IResultHandler, start: typing.Union[datetime.datetime, datetime.date], end: typing.Union[datetime.datetime, datetime.date], end_of_day_delta: datetime.timedelta, current_utc_time: typing.Optional[datetime.datetime] = None) -> QuantConnect.Scheduling.ScheduledEvent:
        """
        Creates a new ScheduledEvent that will fire before market close by the specified time
        
        This method is deprecated. It will generate ScheduledEvents for the deprecated IAlgorithm.OnEndOfDay()
        
        :param algorithm: The algorithm instance the event is fo
        :param result_handler: The result handler, used to communicate run time errors
        :param start: The date to start the events
        :param end: The date to end the events
        :param end_of_day_delta: The time difference between the market close and the event, positive time will fire before market close
        :param current_utc_time: Specfies the current time in UTC, before which, no events will be scheduled. Specify null to skip this filter.
        :returns: The new ScheduledEvent that will fire near market close each tradeable dat.
        """
        warnings.warn("This method is deprecated. It will generate ScheduledEvents for the deprecated IAlgorithm.OnEndOfDay()", DeprecationWarning)

    @staticmethod
    def every_day_at(name: str, dates: System.Collections.Generic.IEnumerable[datetime.datetime], time_of_day: datetime.timedelta, callback: typing.Callable[[str, datetime.datetime], None], current_utc_time: typing.Optional[datetime.datetime] = None) -> QuantConnect.Scheduling.ScheduledEvent:
        """
        Creates a new ScheduledEvent that will fire at the specified  for every day in
        
        :param name: An identifier for this event
        :param dates: The dates to set events for at the specified time. These act as a base time to which the  is added to, that is, the implementation does not use .Date before the addition
        :param time_of_day: The time each tradeable date to fire the event
        :param callback: The delegate to call when an event fires
        :param current_utc_time: Specfies the current time in UTC, before which, no events will be scheduled. Specify null to skip this filter.
        :returns: A new ScheduledEvent instance that fires events each tradeable day from the start to the finish at the specified time.
        """
        ...

    @staticmethod
    def every_security_end_of_day(algorithm: QuantConnect.Interfaces.IAlgorithm, result_handler: QuantConnect.Lean.Engine.Results.IResultHandler, security: QuantConnect.Securities.Security, start: typing.Union[datetime.datetime, datetime.date], end: typing.Union[datetime.datetime, datetime.date], end_of_day_delta: datetime.timedelta, current_utc_time: typing.Optional[datetime.datetime] = None) -> QuantConnect.Scheduling.ScheduledEvent:
        """
        Creates a new ScheduledEvent that will fire before market close by the specified time
        
        :param algorithm: The algorithm instance the event is fo
        :param result_handler: The result handler, used to communicate run time errors
        :param security: The security used for defining tradeable dates
        :param start: The first date for the events
        :param end: The date to end the events
        :param end_of_day_delta: The time difference between the market close and the event, positive time will fire before market close
        :param current_utc_time: Specfies the current time in UTC, before which, no events will be scheduled. Specify null to skip this filter.
        :returns: The new ScheduledEvent that will fire near market close each tradeable dat.
        """
        ...





from typing import overload
from enum import Enum
import abc

import System
import System.Runtime.ConstrainedExecution


class Consistency(Enum):
    """Obsoletions.ConstrainedExecutionRegionMessage"""

    MAY_CORRUPT_PROCESS = 0

    MAY_CORRUPT_APP_DOMAIN = 1

    MAY_CORRUPT_INSTANCE = 2

    WILL_NOT_CORRUPT_STATE = 3


class Cer(Enum):
    """Obsoletions.ConstrainedExecutionRegionMessage"""

    NONE = 0

    MAY_FAIL = 1

    SUCCESS = 2


class PrePrepareMethodAttribute(System.Attribute):
    """Obsoletions.ConstrainedExecutionRegionMessage"""

    def __init__(self) -> None:
        ...


class CriticalFinalizerObject(System.Object, metaclass=abc.ABCMeta):
    """Ensures that all finalization code in derived classes is marked as critical."""

    def __init__(self) -> None:
        """This method is protected."""
        ...


class ReliabilityContractAttribute(System.Attribute):
    """
    Defines a contract for reliability between the author of some code, and the developers who have a dependency on that code.
    
    Obsoletions.ConstrainedExecutionRegionMessage
    """

    @property
    def consistency_guarantee(self) -> System.Runtime.ConstrainedExecution.Consistency:
        ...

    @property
    def cer(self) -> System.Runtime.ConstrainedExecution.Cer:
        ...

    def __init__(self, consistencyGuarantee: System.Runtime.ConstrainedExecution.Consistency, cer: System.Runtime.ConstrainedExecution.Cer) -> None:
        ...





from typing import overload
from enum import Enum
import System
import System.Diagnostics.CodeAnalysis


class ExcludeFromCodeCoverageAttribute(System.Attribute):
    """This class has no documentation."""

    @property
    def justification(self) -> str:
        """Gets or sets the justification for excluding the member from code coverage."""
        ...

    @property.setter
    def justification(self, value: str) -> None:
        ...

    def __init__(self) -> None:
        ...


class SuppressMessageAttribute(System.Attribute):
    """Suppresses reporting of a specific code analysis rule violation, allowing multiple suppressions on a single code artifact. Does not apply to compiler diagnostics."""

    @property
    def category(self) -> str:
        ...

    @property
    def check_id(self) -> str:
        ...

    @property
    def scope(self) -> str:
        ...

    @property.setter
    def scope(self, value: str) -> None:
        ...

    @property
    def target(self) -> str:
        ...

    @property.setter
    def target(self, value: str) -> None:
        ...

    @property
    def message_id(self) -> str:
        ...

    @property.setter
    def message_id(self, value: str) -> None:
        ...

    @property
    def justification(self) -> str:
        ...

    @property.setter
    def justification(self, value: str) -> None:
        ...

    def __init__(self, category: str, checkId: str) -> None:
        ...


class ConstantExpectedAttribute(System.Attribute):
    """Indicates that the specified method parameter expects a constant."""

    @property
    def min(self) -> System.Object:
        """Indicates the minimum bound of the expected constant, inclusive."""
        ...

    @property.setter
    def min(self, value: System.Object) -> None:
        ...

    @property
    def max(self) -> System.Object:
        """Indicates the maximum bound of the expected constant, inclusive."""
        ...

    @property.setter
    def max(self, value: System.Object) -> None:
        ...


class UnscopedRefAttribute(System.Attribute):
    """Used to indicate a byref escapes and is not scoped."""

    def __init__(self) -> None:
        """Initializes a new instance of the UnscopedRefAttribute class."""
        ...





from typing import overload
from enum import Enum
import typing

import QuantConnect
import QuantConnect.Orders
import QuantConnect.Securities
import QuantConnect.Securities.CryptoFuture
import System


class CryptoFutureExchange(QuantConnect.Securities.SecurityExchange):
    """Crypto future exchange class - information and helper tools for Crypto future exchange properties"""

    @overload
    def __init__(self, market: str) -> None:
        """
        Initializes a new instance of the CryptoFutureExchange class using market hours
        derived from the market-hours-database for the Crypto future market
        """
        ...

    @overload
    def __init__(self, exchangeHours: QuantConnect.Securities.SecurityExchangeHours) -> None:
        """
        Initializes a new instance of the CryptoFutureExchange class using the specified
        exchange hours to determine open/close times
        
        :param exchangeHours: Contains the weekly exchange schedule plus holidays
        """
        ...


class BinanceFutureMarginInterestRateModel(System.Object, QuantConnect.Securities.IMarginInterestRateModel):
    """The responsability of this model is to apply future funding rate cash flows to the portfolio based on open positions"""

    def apply_margin_interest_rate(self, margin_interest_rate_parameters: QuantConnect.Securities.MarginInterestRateParameters) -> None:
        """
        Apply margin interest rates to the portfolio
        
        :param margin_interest_rate_parameters: The parameters to use
        """
        ...


class BybitFutureMarginInterestRateModel(QuantConnect.Securities.CryptoFuture.BinanceFutureMarginInterestRateModel):
    """The responsibility of this model is to apply future funding rate cash flows to the portfolio based on open positions"""


class CryptoFutureMarginModel(QuantConnect.Securities.SecurityMarginModel):
    """The crypto future margin model which supports both Coin and USDT futures"""

    def __init__(self, leverage: float = 25, maintenanceMarginRate: float = 0.05, maintenanceAmount: float = 0) -> None:
        """
        Creates a new instance
        
        :param leverage: The leverage to use, used on initial margin requirements, default 25x
        :param maintenanceMarginRate: The maintenance margin rate, default 5%
        :param maintenanceAmount: The maintenance amount which will reduce maintenance margin requirements, default 0
        """
        ...

    def get_initial_margin_requirement(self, parameters: QuantConnect.Securities.InitialMarginParameters) -> QuantConnect.Securities.InitialMargin:
        """
        The margin that must be held in order to increase the position by the provided quantity
        
        :param parameters: An object containing the security and quantity of shares
        :returns: The initial margin required for the option (i.e. the equity required to enter a position for this option).
        """
        ...

    def get_maintenance_margin(self, parameters: QuantConnect.Securities.MaintenanceMarginParameters) -> QuantConnect.Securities.MaintenanceMargin:
        """
        Gets the margin currently alloted to the specified holding.
        
        :param parameters: An object containing the security
        :returns: The maintenance margin required for the option.
        """
        ...

    def get_margin_remaining(self, portfolio: QuantConnect.Securities.SecurityPortfolioManager, security: QuantConnect.Securities.Security, direction: QuantConnect.Orders.OrderDirection) -> float:
        """
        Gets the margin cash available for a trade
        
        This method is protected.
        
        :param portfolio: The algorithm's portfolio
        :param security: The security to be traded
        :param direction: The direction of the trade
        :returns: The margin available for the trade.
        """
        ...


class CryptoFuture(QuantConnect.Securities.Security, QuantConnect.Securities.IBaseCurrencySymbol):
    """Crypto Future Security Object Implementation for Crypto Future Assets"""

    @property
    def base_currency(self) -> QuantConnect.Securities.Cash:
        """Gets the currency acquired by going long this currency pair"""
        ...

    def __init__(self, symbol: typing.Union[QuantConnect.Symbol, str], exchangeHours: QuantConnect.Securities.SecurityExchangeHours, quoteCurrency: QuantConnect.Securities.Cash, baseCurrency: QuantConnect.Securities.Cash, symbolProperties: QuantConnect.Securities.SymbolProperties, currencyConverter: QuantConnect.Securities.ICurrencyConverter, registeredTypes: QuantConnect.Securities.IRegisteredSecurityDataTypesProvider, cache: QuantConnect.Securities.SecurityCache) -> None:
        """
        Constructor for the Crypto Future security
        
        :param symbol: The symbol
        :param exchangeHours: Defines the hours this exchange is open
        :param quoteCurrency: The cash object that represent the quote currency
        :param baseCurrency: The cash object that represent the base currency
        :param symbolProperties: The symbol properties for this security
        :param currencyConverter: Currency converter used to convert CashAmount instances into units of the account currency
        :param registeredTypes: Provides all data types registered in the algorithm
        :param cache: The security cache
        """
        ...

    def is_crypto_coin_future(self) -> bool:
        """
        Checks whether the security is a crypto coin future
        
        :returns: True if the security is a crypto coin future.
        """
        ...


class CryptoFutureHolding(QuantConnect.Securities.SecurityHolding):
    """Crypto Future holdings implementation of the base securities class"""

    def __init__(self, security: QuantConnect.Securities.Security, currencyConverter: QuantConnect.Securities.ICurrencyConverter) -> None:
        """
        Crypto Future Holding Class constructor
        
        :param security: The crypto future security being held
        :param currencyConverter: A currency converter instance
        """
        ...

    def get_quantity_value(self, quantity: float, price: float) -> QuantConnect.Securities.ConvertibleCashAmount:
        """
        Gets the total value of the specified  of shares of this security
        in the account currency
        
        :param quantity: The quantity of shares
        :param price: The current price
        :returns: The value of the quantity of shares in the account currency.
        """
        ...





from typing import overload
from enum import Enum
import abc

import System
import System.Security.Principal


class IIdentity(metaclass=abc.ABCMeta):
    """This class has no documentation."""

    @property
    @abc.abstractmethod
    def name(self) -> str:
        ...

    @property
    @abc.abstractmethod
    def authentication_type(self) -> str:
        ...

    @property
    @abc.abstractmethod
    def is_authenticated(self) -> bool:
        ...


class PrincipalPolicy(Enum):
    """This class has no documentation."""

    UNAUTHENTICATED_PRINCIPAL = 0

    NO_PRINCIPAL = 1

    WINDOWS_PRINCIPAL = 2


class TokenImpersonationLevel(Enum):
    """This class has no documentation."""

    NONE = 0

    ANONYMOUS = 1

    IDENTIFICATION = 2

    IMPERSONATION = 3

    DELEGATION = 4


class IPrincipal(metaclass=abc.ABCMeta):
    """This class has no documentation."""

    @property
    @abc.abstractmethod
    def identity(self) -> System.Security.Principal.IIdentity:
        ...

    def is_in_role(self, role: str) -> bool:
        ...


